"""Generate rules to use in pickaxe runs."""
from bisect import bisect_right
from io import StringIO
from os import remove
from pathlib import Path
from typing import List, Tuple

import pandas as pd


pwd = Path(__file__).parent

pattern_dictionary = {
    "aromatic": r":\[|\]:",
    "aromatic_oxygen": r"^\[#6:\d+\]:|\[#6:\d+\]:|\[#6:\d+\]\d:|\[#6:\d+\]:\d",
    "carbonyl": r"=\[#8:\d\]|\[#8:\d\]=",
    "nitrogen": r"\[#7:\d+\]",
    "oxygen": r"\[#8:\d+\]",
    "fluorine": r"\[#9:\d+\]",
    "phosphorus": r"\[#15:\d+\]",
    "sulfur": r"\[#16:\d+\]",
    "chlorine": r"\[#17:\d+\]",
    "bromine": r"\[#35:\d+\]",
    "iodine": r"\[#53:\d+\]",
    "halogen": r"\[#(9|17|35|53):\d+\]",
}


def metacyc_generalized(
    n_rules: int = None,
    fraction_coverage: float = None,
    anaerobic: float = False,
    include_containing: List[str] = None,
    exclude_containing: List[str] = None,
    **kwargs,
) -> Tuple[StringIO, StringIO, str]:
    """Generate generalize metacyc rule subsets.

    Generate subsets of the metacyc generalized reaction opreators by specifying the
    number of rules of the fraction coverage of metacyc desired. Rules are chosen
    in the order of rules that map the most reactions to least. For fractional coverage
    the lowest number of rules that give a coverage less than or equal to the specified
    coverage is given.

    Specific groups can be specified to be excluded or used as well to prune to specific
    rules. This is a two step process:
        1) Select rules by include_containing. If none specified, use all rules.
        2) Remove rules by excluded_containing.

    Parameters
    ----------
    n_rules : int, optional
        Number of rules to use. If excluded rules result in less than specified
        number then all rules are taken, by default None.
    fraction_coverage : float, optional
        The fraction of coverage desired. This may be impossible to reach depending
        on which rules are excluded, by default None.
    anaerobic: float, optional
        Whether to remove oxygen requiring reactions.
    include_containing: List[str], optional
        A list containing features to include. Valid features are:
            - aromatic
            - aromatic_oxygen
            - carbonyl
            - halogen
            - nitrogen
            - oxygen
            - phosphorus
            - sulfur
            - fluorine
            - chlorine
            - bromine
            - iodine
        sending None gives all groups, by default None.
    exclude_containing: List[str], optional
        A list containing features to exclude.
            - aromatic
            - aromatic_oxygen
            - carbonyl
            - halogen
            - nitrogen
            - oxygen
            - phosphorus
            - sulfur
            - fluorine
            - chlorine
            - bromine
            - iodine
        By default None.

    Returns
    -------
    Tuple[StringIO, StringIO, str]
        A tuple containing two streams (reaction rules and cofactor) and
        the rule name.
    """
    # Whether or not to return counts
    return_counts = kwargs.get("return_counts", False)
    return_all = kwargs.get("return_all", False)
    # Metacyc Rules
    reaction_mapping = pwd / Path("data/metacyc_rules/metacyc21_coverage.tsv")
    rules = pwd / Path("data/metacyc_rules/metacyc_generalized_rules.tsv")
    coreactants = pwd / Path("data/metacyc_rules/metacyc_coreactants.tsv")

    # Get dataframe to help select rules
    rule_df = pd.read_csv(reaction_mapping, delimiter="\t", usecols=["Rules"]).rename(
        columns={"Rules": "Name"}
    )

    # Generate CDF for determining fraction coverage
    rule_counts = rule_df.value_counts().rename_axis("Name").reset_index(name="counts")

    # Attach rules to this DF
    rule_df = pd.read_csv(
        rules,
        delimiter="\t"
        # usecols=["Name", "Reactants", "SMARTS", "Prod"]
    )
    removed_df = pd.DataFrame()

    # Filter out any reactions based on filtering
    name_append = ""
    removed_ids = set()
    if anaerobic:
        removed_ids.update(
            rule_df[rule_df["Reactants"].str.contains("^O2|;O2|O2;")].index.to_list()
        )
        # rule_df = rule_df[~rule_df["Reactants"].str.contains("^O2|;O2|O2;")]
        name_append += "_anaerobic"

    if include_containing:
        patterns_to_include = r"|".join(
            pattern_dictionary[feature] for feature in include_containing
        )
        removed_ids.update(
            rule_df[
                ~rule_df["SMARTS"].str.contains(patterns_to_include)
            ].index.to_list()
        )
        # rule_df = rule_df[rule_df["SMARTS"].str.contains(patterns_to_include)]
        name_append += "_with_inclusion"

    if exclude_containing:
        patterns_to_exclude = r"|".join(
            pattern_dictionary[feature] for feature in exclude_containing
        )
        removed_ids.update(
            rule_df[rule_df["SMARTS"].str.contains(patterns_to_exclude)].index.to_list()
        )
        # rule_df = rule_df[~rule_df["SMARTS"].str.contains(patterns_to_exclude)]
        name_append += "_with_exclusion"

    # Reorder dataframe
    rule_ids = rule_df.index.to_list()
    keep_ids = [i for i in rule_ids if i not in removed_ids]
    removed_ids = [i for i in rule_ids if i in removed_ids]

    new_ids = keep_ids + removed_ids

    # Merge with rule counts, saving old ids
    rule_df["index"] = rule_df.index
    rule_df = pd.merge(rule_counts, rule_df, on="Name").set_index("index")

    # TODO why is this happening?
    # Calculate ids to keep, some rules are dropped due to not having reactions
    new_ids = [i for i in new_ids if i in rule_df.index]
    rule_df = rule_df.loc[new_ids]

    # Calculate CDF and drop unwanted rules
    rule_df["cdf"] = rule_df["counts"].cumsum() / rule_df["counts"].sum()

    if return_all:
        rule_df.loc[rule_df.index.isin(keep_ids), "type"] = "keep"
        rule_df.loc[~rule_df.index.isin(keep_ids), "type"] = "drop"
    else:
        rule_df = rule_df[rule_df.index.isin(keep_ids)]
        rule_df = rule_df.reset_index(drop=False)

    # Get regex pattern to search for rules to filter dataframe
    if n_rules:
        # Default to n_rules if all are specified
        # pat = "|".join(list(rule_df.iloc[0:n_rules]["rule"]))
        rule_name = f"Metacyc_generalized_{n_rules}_rules"
    elif fraction_coverage:
        # Calculate CDF
        n_rules = bisect_right(rule_df["cdf"].values, fraction_coverage) + 1
        # n_rules = rule_df["cdf"].sub(fraction_coverage).abs().idxmin() + 1
        # pat = "|".join(list(rule_df.iloc[0:n_rules]["rule"]))
        rule_name = (
            f"Metacyc_generalized_{fraction_coverage}_fraction_coverage".replace(
                ".", ","
            )
        )
    else:
        # pat = "|".join(list(rule_counts["rule"]))
        n_rules = len(rule_df)
        rule_name = "Metacyc_generalized"

    rule_name += name_append

    # get stream to new rules
    new_rules = rule_df.iloc[0:n_rules]
    if not return_counts:
        new_rules = new_rules.drop(columns=["counts"])
        if "cdf" in new_rules.columns:
            new_rules = new_rules.drop(columns=["cdf"])

    rule_stream = StringIO()
    new_rules.to_csv(rule_stream, sep="\t", index=False)
    rule_stream = StringIO(rule_stream.getvalue())

    return rule_stream, coreactants, rule_name


def metacyc_intermediate(
    n_rules: int = None,
    fraction_coverage: float = None,
    anaerobic: float = False,
    include_containing: List[str] = None,
    exclude_containing: List[str] = None,
) -> Tuple[StringIO, StringIO, str]:
    """Generate intermediate metacyc rule subsets.

    Generate subsets of the metacyc intermediate reaction opreators by specifying the
    number of rules of the fraction coverage of metacyc desired. Coverage and number of
    rules are taken from the generalized operators and their intermediate operators are
    chosen.

    Parameters
    ----------
    n_rules : int, optional
        Number of rules to use, by default None.
    fraction_coverage : float, optional
        The fraction of coverage desired, by default None.
    anaerobic: float, optional
        Whether to remove oxygen requiring reactions.
    include_containing: List[str], optional
        A list containing features to include. Valid features are:
            - aromatic
            - aromatic_oxygen
            - carbonyl
            - halogen
            - nitrogen
            - oxygen
            - phosphorus
            - sulfur
            - fluorine
            - chlorine
            - bromine
            - iodine
        sending None gives all groups, by default None.
    exclude_containing: List[str], optional
        A list containing features to exclude. Valid features are:
            - aromatic
            - aromatic_oxygen
            - carbonyl
            - halogen
            - nitrogen
            - oxygen
            - phosphorus
            - sulfur
            - fluorine
            - chlorine
            - bromine
            - iodine
        by default None

    Returns
    -------
    Tuple[StringIO, StringIO, str]
        A tuple containing two streams that contain the reaction rule information and
        the rule name.
    """
    # Metacyc Rules
    rules = pwd / Path("data/metacyc_rules/metacyc_intermediate_rules.tsv")
    # uniprot_rules = pwd / Path(
    #     "data/metacyc_rules/metacyc_intermediate_rules_uniprot.tsv"
    # )
    coreactants = pwd / Path("data/metacyc_rules/metacyc_coreactants.tsv")

    # Get intermediate dataframe and sort by reactions mapped
    rule_df = pd.read_csv(rules, delimiter="\t")
    # rule_df["counts"] = rule_df.Comments.map(lambda s: len(s.split(";")))
    total_rxns = rule_df.counts.sum()

    # Filter DF to only desired compounds using general rules pattern matching
    general_rule_stream, _, _ = metacyc_generalized(
        n_rules=None,
        fraction_coverage=1,
        anaerobic=anaerobic,
        include_containing=include_containing,
        exclude_containing=exclude_containing,
        return_counts=True,
    )
    general_rule_df = pd.read_csv(
        general_rule_stream,
        delimiter="\t",
        usecols=["Name", "Reactants", "SMARTS", "Products", "counts", "Comments"],
    )
    valid_rules_pattern = "|".join(rule for rule in general_rule_df["Name"].values)
    rule_df = rule_df[rule_df["Name"].str.contains(valid_rules_pattern)]

    # Some generalized have no intermediate, bring those in
    # missing_rules = set(general_rule_df.Name) - set(
    #     v.split("_")[0] for v in rule_df["Name"].values
    # )
    # if missing_rules:
    #     missing_df = general_rule_df[
    #         general_rule_df["Name"].str.contains("|".join(missing_rules))
    #     ]
    #     rule_df = rule_df.append(missing_df)

    # Generate
    # CDF for determining fraction coverage
    rule_df = rule_df.sort_values(by="counts", ascending=False)
    rule_df["cdf"] = rule_df["counts"].cumsum() / total_rxns

    # saved file to do this instead
    # # Change Comments to be uniprot
    # uniprot_df = pd.read_csv(
    #     uniprot_rules, delimiter="\t", usecols=["Name", "Comments"]
    # )
    # uniprot_df = uniprot_df.rename(columns={"Comments": "Uniprot"})
    # # rule_df = rule_df.drop(columns=["Comments"])
    # rule_df = pd.merge(rule_df, uniprot_df, on="Name", how="left")

    # Filter out any reactions based on filtering
    name_append = ""
    if anaerobic:
        name_append += "_anaerobic"

    if include_containing:
        name_append += "_with_inclusion"

    if exclude_containing:
        name_append += "_with_exclusion"

    # Get regex pattern to search for rules to filter dataframe
    if n_rules:
        # Default to n_rules if all are specified
        rule_name = f"Metacyc_intermediate_{n_rules}_rules"
    elif fraction_coverage:
        # Calculate CDF
        n_rules = bisect_right(rule_df["cdf"].values, fraction_coverage) + 1
        rule_name = (
            f"Metacyc_intermediate_{fraction_coverage}_fraction_coverage".replace(
                ".", ","
            )
        )
    else:
        # pat = "|".join(list(rule_counts["rule"]))
        n_rules = len(rule_df)
        rule_name = "Metacyc_intermediate"

    # get stream to new rules
    rule_name += name_append

    # get stream to new rules
    new_rules = rule_df.iloc[0:n_rules]
    if "cdf" in new_rules.columns:
        new_rules = new_rules.drop(columns=["cdf"])
    if "counts" in new_rules.columns:
        new_rules = new_rules.drop(columns=["counts"])
    rule_stream = StringIO()
    new_rules.to_csv(rule_stream, sep="\t", index=False)
    rule_stream = StringIO(rule_stream.getvalue())

    return rule_stream, coreactants, rule_name


def BNICE() -> Tuple[Path, Path, str]:
    """Generate BNICE rules.

    Generate the original BNICE rules that were use before the improved MetaCyc rules
    were generated.

    Returns
    -------
    Tuple[Path, Path, str]
        The path to the rules and coreactants and the rule name.
    """
    rules = pwd / Path("data/original_rules/EnzymaticReactionRules.tsv")
    coreactants = pwd / Path("data/original_rules/EnzymaticCoreactants.tsv")
    rule_name = "BNICE"

    return rules, coreactants, rule_name


import seaborn as sns


sns.scatterplot()



"""Tests for metabolomics.py using pytest."""

from pathlib import Path

import pymongo
import pytest
from pymongo.errors import ServerSelectionTimeoutError

from minedatabase.metabolomics import (
    MetabolomicsDataset,
    Peak,
    dot_product,
    jaccard,
    read_mgf,
    read_msp,
    read_mzxml,
    spectra_download,
)


try:
    client = pymongo.MongoClient(ServerSelectionTimeoutMS=20)
    client.server_info()
    del client
    is_mongo = True
except ServerSelectionTimeoutError as err:
    is_mongo = False
valid_db = pytest.mark.skipif(not is_mongo, reason="No MongoDB Connection")

file_path = Path(__file__)
file_dir = file_path.parent

DATA_DIR = (file_dir / "../data/").resolve()

# -------------------------------- Fixtures --------------------------------- #
@pytest.fixture
def known_peaks():
    """Return a list of Peak instances with data (identified)."""
    peak1 = Peak(
        name="Test1Known",
        r_time=5.00,
        mz=867.1391,
        charge="+",
        inchi_key="IRPOHFRNKHKIQA-UHFFFAOYSA-N",
    )
    peak2 = Peak(
        name="Test2Known",
        r_time=8.00,
        mz=260.0297,
        charge="-",
        inchi_key="HXXFSFRBOHSIMQ-FPRJBGLDSA-N",
    )
    return [peak1, peak2]


@pytest.fixture
def unknown_peaks():
    """Return a list of Peak instances with data (unidentified)."""
    peak1 = Peak(
        name="Test1Unknown",
        r_time=3.00,
        mz=427.0294,
        charge="+",
        ms2=[(10, 110), (20, 320), (25, 5)],
    )
    peak2 = Peak(
        name="Test2Unknown",
        r_time=1.50,
        mz=180.0000,
        charge="-",
        ms2=[(10, 105), (20, 50), (25, 90)],
    )
    return [peak1, peak2]


@pytest.fixture
def native_cpd_set():
    """Return a set of native compound IDs."""
    return set(
        [
            "C03e0b10e6490ce79a7b88cb0c4e17c2bf6204352",
            "C189c6f78587772ccbd0c2b9e118cf88ff35c316a",
            "C90ba1f8d4d84f6305539f6d05a74497d4d5dfe06",
        ]
    )


@pytest.fixture
def metabolomics_dataset(known_peaks, unknown_peaks, native_cpd_set):
    """Default instance of MetabolomicsDataset."""
    met_dataset = MetabolomicsDataset(
        name="Test Metabolomics Dataset",
        adducts=["[M-H]-", "[M+H]+"],
        known_peaks=known_peaks,
        unknown_peaks=unknown_peaks,
        native_set=native_cpd_set,
        ppm=False,
        tolerance=0.002,
        halogens=True,
        verbose=True,
    )
    return met_dataset


# ------------------------ MetabolomicsDataset tests ------------------------ #
def test_metabolomics_dataset_adducts(metabolomics_dataset):
    """Test adducts of a metabolomics dataset.
    GIVEN an instance MetabolomicsDataset
    WHEN an it is initialized
    THEN make sure it has the correct adducts
    """
    assert metabolomics_dataset.pos_adducts == [("[M+H]+", 1.0, 1.007276)]
    assert metabolomics_dataset.neg_adducts == [("[M-H]-", 1.0, -1.007276)]


def test_metabolomics_dataset_str_representation(metabolomics_dataset):
    """Test MetabolomicsDataset object string representation.
    GIVEN an instance of MetabolomicsDataset
    WHEN the string representation is accessed (e.g. using str() or print())
    THEN make sure the metabolomics dataset name is returned
    """
    assert str(metabolomics_dataset) == metabolomics_dataset.name


def test_metabolomics_dataset_enumerate_possible_masses(metabolomics_dataset):
    """Make sure possible masses and ranges are calculated properly.
    GIVEN an instance of MetabolomicsDataset
    WHEN enumerate_possible_masses is used to get mass ranges
    THEN make sure those mass ranges are correctly calculated
    """
    metabolomics_dataset.enumerate_possible_masses(tolerance=0.001)
    # possible masses are each peak mass +/- each adduct
    possible_masses = {"+": set([426.022124]), "-": set([181.007276])}
    assert metabolomics_dataset.possible_masses == possible_masses
    # possible ranges are possible masses +/- tolerance
    possible_ranges = {
        "+": [(426.02112400000004, 426.023124, "Test1Unknown", "[M+H]+")],
        "-": [(181.00627599999999, 181.008276, "Test2Unknown", "[M-H]-")],
    }
    assert metabolomics_dataset.possible_ranges == possible_ranges


def test_metabolomics_dataset_get_rt(metabolomics_dataset):
    """Make sure we can get a Peak's retention time from its ID.
    GIVEN an instance of MetabolomicsDataset
    WHEN the retention time for a given peak in that dataset is requested
    THEN make sure the correct retention time is returned
    """
    peak1_actual_rt = 5.00
    peak2_actual_rt = 8.00

    assert metabolomics_dataset.get_rt("Test1Known") == peak1_actual_rt
    assert metabolomics_dataset.get_rt("Test2Known") == peak2_actual_rt
    assert metabolomics_dataset.get_rt("InvalidID") is None


# TODO Jon fix me!
@valid_db
def test_metabolomics_dataset_find_db_hits(test_db, metabolomics_dataset):
    """Search for expected metaoblomics hits in test database
    GIVEN a MINE database and metabolomics dataset
    WHEN that database is searched against a specific peak in that dataset
    THEN make sure the correct hits are returned
    """
    peak = Peak(name="test", r_time=1, mz=261.0369946, charge="+")
    adducts = [("[M+H]+", 1, 1.007276), ("[M-H]-", 1, -1.007276)]
    metabolomics_dataset.find_db_hits(peak, test_db, adducts)
    assert len(peak.isomers) == 1


@valid_db
def test_metabolomics_dataset_annotate_peaks(test_db, metabolomics_dataset):
    """Uses find_db_hits to try to annotate all unknown peaks in dataset
    GIVEN a metabolomics dataset and MINE db
    WHEN trying to find hits for every unknown peak in that dataset
    THEN make sure all peaks get the correct # of hits
    """
    if not test_db.compounds.find_one({"_id": "mass_test"}):
        test_db.compounds.insert_one(
            {
                "_id": "mass_test",
                "Mass": 181.007276,
                "Charge": 0,
                "Formula": "C6H12O2N",
                "Generation": 1,
            }
        )
    metabolomics_dataset.annotate_peaks(test_db)
    for peak in metabolomics_dataset.unknown_peaks:
        if peak.name == "Test1Unknown":
            assert len(peak.isomers) == 0
        elif peak.name == "Test2Unknown":
            assert len(peak.isomers) == 1


# -------------------------- Scoring function tests ------------------------- #
def test_dot_product():
    """Test dot product calculation.
    GIVEN two spectra
    WHEN their dot product is calculated
    THEN make sure it is the correct value
    """
    x = [(10, 100), (20, 300), (25, 10)]
    y = [(10.011, 10), (20.009, 300), (25, 100)]
    epsilon = 0.01

    calc_dot_product = dot_product(x, y, epsilon=epsilon)
    assert round(calc_dot_product, 5) == 0.90909


def test_jaccard():
    """Test jaccard index calculation.
    GIVEN two spectra
    WHEN their jaccard index is calculated
    THEN make sure it is the correct value
    """
    x = [(10, 100), (20, 300), (25, 10)]
    y = [(10.011, 10), (20.009, 300), (25, 100)]
    epsilon = 0.01

    calc_jaccard = jaccard(x, y, epsilon=epsilon)
    assert calc_jaccard == 0.5


# -------------------------------- Peak tests ------------------------------- #
def test_peak_str_representation(known_peaks):
    """Test Peak object string representation.
    GIVEN an instance of Peak
    WHEN the string representation is accessed (e.g. using str() or print())
    THEN make sure the peak name is returned
    """
    for peak in known_peaks:
        assert str(peak) == peak.name


def test_peak_score_isomers(unknown_peaks):
    """Test score_isomers method of Peak object.
    GIVEN an instance of Peak
    WHEN isomers are scored against mass spectra data
    THEN make sure they are in the correct order
    """
    # In reality pos and neg spectra would be different, but this is just for testing
    isomer1 = {'Spectra':
        {
            "Positive": {"20V": [(10, 100), (20, 300), (25, 10)]},
            "Negative": {"20V": [(10, 100), (20, 300), (25, 10)]},
        }
    }
    isomer2 = {'Spectra':
        {
            "Positive": {"20V": [(10, 10), (20, 55), (25, 100)]},
            "Negative": {"20V": [(10, 10), (20, 55), (25, 100)]},
        }
    }

    unknown_peak1, unknown_peak2 = unknown_peaks
    unknown_peak1.isomers = [isomer1, isomer2]
    unknown_peak2.isomers = [isomer1, isomer2]

    # Isomer order should get switched for peak2 but not for peak1
    unknown_peak1.score_isomers()
    unknown_peak2.score_isomers()

    assert unknown_peak1.isomers == [isomer1, isomer2]
    assert unknown_peak2.isomers == [isomer2, isomer1]


# --------------------- Metabolomics File Parsing Tests --------------------- #


def test_read_mgf():
    """Test reading peak data from an MGF metabolomics file.
    GIVEN an MGF file
    WHEN that MGF file is parsed into a list of Peak objects
    THEN make sure those Peak objects are correct
    """
    mgf_path = DATA_DIR / "test_metabolomics/test.mgf"
    with open(mgf_path, "r") as infile:
        mgf_data = infile.read()

    peaks = read_mgf(mgf_data, charge="+")
    test_peak1 = peaks[0]
    test_peak2 = peaks[1]

    assert test_peak1.charge == "+"
    assert test_peak1.mz == 100.5
    assert test_peak1.ms2peaks == [(40.4, 2.3), (50.3, 12.3), (102.1, 10.0)]
    assert test_peak1.r_time == 420

    assert test_peak2.charge == "+"
    assert test_peak2.mz == 131.0
    assert test_peak2.ms2peaks == [(93.8, 12.4), (250.1, 1000.2)]
    assert test_peak2.r_time == 305


def test_read_msp():
    """Test reading peak data from an MSP metabolomics file.
    GIVEN an MSP file
    WHEN that MSP file is parsed into a list of Peak objects
    THEN make sure those Peak objects are correct
    """
    msp_path = DATA_DIR / "test_metabolomics/test.msp"
    with open(msp_path, "r") as infile:
        msp_data = infile.read()

    peaks = read_msp(msp_data, charge="+")
    test_peak1 = peaks[0]
    test_peak2 = peaks[1]

    assert test_peak1.charge == "+"
    assert test_peak1.mz == 100.5
    assert test_peak1.ms2peaks == [(40.4, 2.3), (50.3, 12.3), (102.1, 10.0)]
    assert test_peak1.r_time == 420

    assert test_peak2.charge == "+"
    assert test_peak2.mz == 131.0
    assert test_peak2.ms2peaks == [(93.8, 12.4), (250.1, 1000.2)]
    assert test_peak2.r_time == 305


def test_read_mzxml():
    """Test reading peak data from an mzXML metabolomics file.
    GIVEN an mzXML file
    WHEN that mzXML file is parsed into a list of Peak objects
    THEN make sure those Peak objects are correct
    """
    mzxml_path = DATA_DIR / "test_metabolomics/test.mzXML"
    with open(mzxml_path, "r") as infile:
        mzxml_data = infile.read()

    peaks = read_mzxml(mzxml_data, charge="+")
    test_peak1 = peaks[0]
    test_peak2 = peaks[1]

    assert test_peak1.charge == "+"
    assert test_peak1.mz == 100.5
    assert test_peak1.r_time == 420

    assert test_peak2.charge == "+"
    assert test_peak2.mz == 131.0
    assert test_peak2.r_time == 305


# ----------------------------- Spectra Download ---------------------------- #


@valid_db
def test_spectra_download(test_db):
    """Test download of spectra from MINE database.
    GIVEN a MINE database
    WHEN spectra are downloaded in string format
    THEN make sure the downloaded spectra is correct
    """
    if not test_db.compounds.find_one({"_id": "ms2_test"}):
        test_spectra = {
            "10V": [(1, 2), (3, 4), (5, 6)],
            "20V": [(6, 7)],
            "40V": [(100, 1000), (250, 200)],
        }
        test_db.compounds.insert_one(
            {
                "_id": "ms2_test",
                "Mass": 260,
                "Charge": 0,
                "Formula": "C12H12O2N",
                "Generation": 1,
                "Pos_CFM_spectra": test_spectra,
            }
        )
    cpd_query = '{"_id": "ms2_test"}'
    spectra = spectra_download(test_db, cpd_query)

    spectra_path = DATA_DIR / "test_metabolomics/test_spectra.txt"

    with open(spectra_path, "r") as infile:
        spectra_data = infile.read()

    assert spectra == spectra_data



from typing import Callable, List

from minedatabase.filters.base_filter import Filter
from minedatabase.pickaxe import Pickaxe


"""Definitions of filters for pickaxe.

Use this module to define your own filter classes. All filter classes must
subclass Filter. See Filter docstring for more information.

To use any filter, import it in pickaxe_run.py, initialize it, and append it
to the .filters property of your pickaxe object.
"""

import copy
import multiprocessing
import time
from functools import partial
from typing import Callable, List, Set, Tuple

import numpy as np
import pandas as pd
import rdkit.rdBase as rkrb
import rdkit.RDLogger as rkl
from rdkit import DataStructs
from rdkit.Chem import AllChem, RDKFingerprint
from rdkit.Chem import rdFMCS as mcs
from rdkit.Chem.rdmolfiles import MolFromSmiles
from rdkit.DataStructs import FingerprintSimilarity
from scipy.stats import rv_discrete

from minedatabase.utils import Chunks


logger = rkl.logger()
logger.setLevel(rkl.ERROR)
rkrb.DisableLog("rdApp.error")

###############################################################################
# Tanimoto Sampling Filter


class SimilaritySamplingFilter(Filter):
    """Filter that samples randomly from weighted similarity scores.

    SimilaritySamplingFilter takes a distribution of similarity scores and uses
    inverse CDF sampling to select N compounds for further expansion. Each compound
    is assigned a similarity score that corresponds to the maximum similarity score
    score of the set of similarity scores obtained by comparing that compound to each
    target. These scores can also be weighted by a specified function to bias higher
    or lower similarity scores scores.

    Types of fingerprints:
        1) RDKit (default)
        2) Morgan

    Types of similarity score:
        1) Tanimoto (default)
        2) Dice

    Parameters
    ----------
    sample_size : int
        Number of compounds to sample.
    weight : Callable
        Function to weight the similarity score with.
    fingerprint_method : str
        Method by which to calculate fingerprints. Options are RDKitFingerprint and
        Morgan, by default RDKitFingerprint.
    fingerprint_args : dict
        A dictionary of keyword arguments for the fingerprint functiono, by default None.
    similarity_method : str
        Method by which to calculate the similarity score. Options are Tanimoto and Dice,
        by default Tanimoto.

    Attributes
    ----------
    sample_size : int
        Number of compounds to sample.
    sample_weight : Callable
        Function to weight the similarity score with.
    fingerprint_method : str
        Fingerprint calculation method.
    fingerprint_args : dict
        Arguments for fingerprint method.
    similairty_method : str
        Simlarity calulation method.
    target_fps : List
        List of fingerprints.
    """

    def __init__(
        self,
        sample_size: int,
        weight: Callable = None,
        fingerprint_method: str = None,
        fingerprint_args: str = None,
        similarity_method: str = None,
    ) -> None:
        self._filter_name = "Similarity Sampling Filter"
        self.sample_size = sample_size
        self.sample_weight = weight

        self.fingerprint_method = fingerprint_method or "RDKit"
        self.fingerprint_args = fingerprint_args or {}
        self.similarity_method = similarity_method or "Tanimoto"

        self.target_fps = []

    @property
    def filter_name(self) -> str:
        return self._filter_name

    def _set_target_fps(self, pickaxe: Pickaxe):
        for smiles in pickaxe.target_smiles:
            mol = MolFromSmiles(smiles)
            if self.fingerprint_method == "Morgan":
                fp = AllChem.GetMorganFingerprintAsBitVect(mol, **self.fingerprint_args)
            else:
                fp = RDKFingerprint(mol)

            self.target_fps.append(fp)

    def _pre_print(self) -> None:
        """Print before filtering."""
        print(
            (
                f"Sampling {self.sample_size} Compounds Based on a "
                f"Weighted Similarity Distribution"
            )
        )

    def _post_print(
        self, pickaxe: Pickaxe, n_total: int, n_filtered: int, time_sample: float
    ) -> None:
        """Print after filtering."""
        print(
            (
                f"{n_filtered} of {n_total} "
                "compounds selected after "
                f"Similarity Sampling of generation {pickaxe.generation}"
                f"--took {time.time() - time_sample}s.\n"
            )
        )

    def _choose_items_to_filter(self, pickaxe: Pickaxe, processes: int) -> Set[str]:
        """
        Samples N compounds to expand based on the weighted Similarity distribution.

        Parameters
        ----------
        pickaxe : Pickaxe
            Pickaxe object to filter
        processes : int
            Number of processes to use.
        """

        print(f"Filtering Generation {pickaxe.generation}" " via Similarity Sampling.")

        if not self.target_fps:
            self._set_target_fps(pickaxe)
        if not self.target_fps:
            print("No targets to filter for. Can't expand.")
            return [], []

        # Get compounds eligible for expansion in the current generation
        compounds_to_check = []
        set_unreactive = False

        for cpd in pickaxe.compounds.values():
            # Compounds are in generation and correct type
            if cpd["Generation"] == pickaxe.generation and cpd["Type"] not in [
                "Coreactant",
                "Target Compound",
            ]:

                # Check for targets and only react if terminal
                if pickaxe.react_targets:
                    compounds_to_check.append(cpd)
                else:
                    for t_id in pickaxe.targets:
                        if "C" + t_id[1:] == cpd["_id"]:
                            set_unreactive = True
                            break

                    if set_unreactive:
                        pickaxe.compounds[cpd["_id"]]["Expand"] = False
                        set_unreactive = False
                    else:
                        compounds_to_check.append(cpd)

        # Get compounds to keep
        cpd_info = [(cpd["_id"], cpd["SMILES"]) for cpd in compounds_to_check]

        sampled_ids = self._sample_by_similarity(
            cpd_info,
            self.target_fps,
            self.sample_size,
            min_S=0.15,
            weighting=self.sample_weight,
            max_iter=None,
            processes=processes,
        )

        # TODO remove me
        print("num sampled = ", len(sampled_ids))
        # Get compounds to remove
        ids = set(i[0] for i in cpd_info)
        cpds_remove_set = ids - sampled_ids

        for c_id in cpds_remove_set:
            pickaxe.compounds[c_id]["Expand"] = False

        return cpds_remove_set, []

    def _sample_by_similarity(
        self,
        mol_info: List[Tuple[str, str]],
        t_fp: List[AllChem.RDKFingerprint],
        n_cpds: int = None,
        min_S: float = 0.05,
        weighting: Callable = None,
        max_iter: int = None,
        processes: int = 1,
    ) -> List[str]:
        """Sample compounds by weighted similarity coefficient.

        Use inverse cumulative distrbution function (CDF) sampling to select
        compounds based on a weighted similarity coefficient distribution.

        Parameters
        ----------
        mol_info : List[Tuple[str, str]]
            A list consisting of (compound_id, SMILES).
        t_fp : List[RDKFingerprint]
            Target fingerprints to compare compounds to.
        n_cpds : int, optional
            Number of compounds to select for sampling, by default None.
        min_S : float, optional
            Minimum similarity similarity to be considered for sampling, by default 0.05.
        weighting : Callable, optional
            Function that accepts a similarity score and returns
            a float, by default None.
        max_iter : int, optional
            The maximum number of iterations before regenerating the CDF for sampling
            , by default None.
        processes : int, optional
            Number of processes to use, by default 1.

        Returns
        -------
        List[str]
            The compound ids to expand.
        """

        # Return input if less than desired number of compounds
        if len(mol_info) <= n_cpds:
            ids = set(x[0] for x in mol_info)
            print(
                "-- Number to sample is less than number of compounds. "
                "Returning all compounds."
            )
            return ids

        # Get pandas df and ids
        df = self.gen_df_from_similarity(
            mol_info, t_fp, min_S=min_S, processes=processes
        )
        if len(df) <= n_cpds:
            ids = set(df["_id"])
            print(
                f"-- After filtering by minimum similarity ({min_S}) "
                "number to sample is less than number of compounds. "
                "Returning all compounds."
            )
            return ids

        print("-- Sampling compounds to expand.")
        then = time.time()
        # Get discrete distribution to sample randomly from
        rv, ids = self._gen_rv_from_df(df, weighting=weighting)

        # Sample intervals from rv and get c_id from id
        if max_iter is None:
            max_iter = n_cpds / 10 if n_cpds > 1000 else n_cpds / 2

        chosen_ids = set()
        nCDF = 0

        collisions_before_recalc = n_cpds / 4
        collisions = 0

        while len(chosen_ids) != n_cpds:
            # Collisions merit reset
            if collisions == collisions_before_recalc:
                collisions = 0
                nCDF += 1
                rv, ids = self._gen_rv_from_df(df, chosen_ids, weighting=weighting)

            sampled_idx = ids.iloc[rv.rvs(size=1)[0]]
            if sampled_idx in chosen_ids:
                # collision
                collisions += 1
            else:
                chosen_ids.add(sampled_idx)
        print(
            f"-- Finished sampling in {time.time() - then} s."
            f" Recalculated CDF {nCDF} times."
        )

        return chosen_ids

    def _gen_rv_from_df(
        self,
        df: pd.DataFrame,
        chosen: List = [],
        weighting: Callable = None,
    ) -> rv_discrete:
        """Generate a scipy.rv object to sample from the inverse CDF.

        Parameters
        ----------
        df : pd.DataFrame
            Dataframe containing the data to sample.
        chosen : List, optional
            Compound ids that have already been chosen, by default [].
        weighting : Callable, optional
            Function to weight the Similarity distribution by, by default None.

        Returns
        -------
        rv_discrete
            scipy.rv object to sample from.
        """
        if weighting is None:

            def weighting(T):
                return T ** 4

        # TODO Make more memory efficient... maybe use np directly instead?
        # Could be due to spawn vs fork
        rescale_df = copy.copy(df[~df["_id"].isin(chosen)])
        rescale_df.loc[:, "T_trans"] = rescale_df["T"].map(weighting)
        rescale_df.loc[:, "T_pdf"] = rescale_df["T_trans"] / sum(rescale_df["T_trans"])

        # Generate CDF
        rescale_df.reset_index(inplace=True, drop=True)
        xk = rescale_df.index
        pk = rescale_df["T_pdf"]
        rv = rv_discrete(values=(xk, pk))
        ids = rescale_df["_id"]

        del rescale_df

        return rv, ids

    def gen_df_from_similarity(
        self,
        mol_info: List[Tuple[str, str]],
        t_fp: List[AllChem.RDKFingerprint],
        min_S: float = 0.05,
        processes: int = 1,
    ) -> pd.DataFrame:
        """Generate a dataframe from similarity

        Parameters
        ----------
        mol_info : List[Tuple[str, str]]
            A list consisting of (compound_id, SMILES).
        t_fp : List[RDKFingerprint]
            Target fingerprints to compare compounds to.
        min_S : float, optional
            Minimum similarity similarity to be considered for sampling, by default 0.05.
        processes : int, optional
            Number of processes to use, by default 1.
        """

        then = time.time()
        print("-- Calculating Fingerprints and Similarity Values.")
        # target fingerprint dataframe
        t_df = pd.DataFrame(t_fp, columns=["fp"])

        # Calculate similarity for each compound and drop T < min_S
        partial_T_calc = partial(
            _calc_max_S,
            t_df,
            min_S,
            self.fingerprint_method,
            self.fingerprint_args,
            self.similarity_method,
        )

        df = pd.DataFrame()
        for mol_chunk in Chunks(mol_info, 10000):
            # Construct targets to sample df
            temp_df = pd.DataFrame(mol_chunk, columns=["_id", "SMILES"])
            df = df.append(_parallelize_dataframe(temp_df, partial_T_calc, processes))

        # Reset index for CDF calculation
        df.reset_index(inplace=True, drop=True)
        print(f"-- Completed Similarity Calculation in {time.time() - then}")

        return df


def _parallelize_dataframe(
    df: pd.DataFrame,
    func: Callable,
    processes: int = 1,
) -> pd.DataFrame:
    """Parallelize mapping a function to a dataframe.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe to apply function to.
    func : Callable
        Function to map onto dataframe.
    processes : int
        Number of processes to use, by default 1.

    Returns
    -------
    df : pd.DataFrame
        New dataframe after having function applied to it in parallel.
    """
    # Require minimum number of compounds to parallelize
    if len(df) <= processes * 4:
        processes = 1

    if processes > 1:
        df_split = np.array_split(df, processes)
        pool = multiprocessing.Pool(processes)
        df = pd.concat(pool.map(func, df_split))
        pool.close()
        pool.join()
    else:
        df = func(df)
    return df


def _calc_max_S(
    t_df: pd.DataFrame,
    min_S: float,
    fingerprint_method: str,
    fingerprint_args: str,
    similarity_method: str,
    df: pd.DataFrame,
) -> pd.DataFrame:
    """Calculate maximum similarity.

    Generate the similarity to use to generate the PMF to sample from.
    For each compound a list of similarity values are obtained by a generated
    compound to every target compound and the max is taken.

    Parameters
    ----------
    t_df : pd.Dataframe
        Dataframe containing the target fingerprints.
    min_S : float
        The minimum similarity similarity score needed to consider a compound.
    fingerprint_method: str
        Which fingerprint method to use, suppoorts RDKit and Morgan, by default RDKit.
    fingerprint_args: dict
        Keyword arguments to pass to fingerprint function, by default empty dict.
    similarity_method: str
        Which similarity method to use. Supports Tanimoto and Dice.
    df : pd.DataFrame
        Dataframe to calculate the max similarity for.

    Returns
    -------
    df : pd.Dataframe
        New dataframe with max similarity values calculated.
    """

    def fingerprint(fingerprint_method, keyword_dict, smi):
        mol = AllChem.MolFromSmiles(smi)
        if fingerprint_method == "Morgan":
            return AllChem.GetMorganFingerprintAsBitVect(mol, **keyword_dict)
        else:
            return RDKFingerprint(mol)

    def calc_similarity(similarity_method, fp1, fp2):
        if similarity_method == "dice":
            return FingerprintSimilarity(fp1, fp2, metric=DataStructs.DiceSimilarity)
        else:
            return FingerprintSimilarity(fp1, fp2)

    fingerprint_partial = partial(fingerprint, fingerprint_method, fingerprint_args)

    df["fp"] = df["SMILES"].map(fingerprint_partial)

    df["T"] = None
    fp = None
    for i in range(len(df)):
        fp = df["fp"].iloc[i]
        df["T"].iloc[i] = max(
            t_df["fp"].map(lambda x: calc_similarity(similarity_method, x, fp))
        )
    # Filter out low similarity
    df = df[df["T"] > min_S]

    return df


# End similarity Sampling Filter
###############################################################################

###############################################################################
# Cutoff filters -- e.g. similarity, MCS metric, and molecular weight


class SimilarityFilter(Filter):
    """A filter that uses a similarity score to determine compounds to expand.

    SimilarityFilter applies a strict cutoff to to the similarity score of
    compounds to determine which compounds to expand.

    Types of fingerprints:
        1) RDKit (default)
        2) Morgan

    Types of similarity score:
        1) Tanimoto (default)
        2) Dice

    Parameters
    ----------
    crit_similarity : float
        The similarity similarity score threshold.
    increasing_similarity : bool
        Whether or not to only keep compounds whos similarity score is higher than its
        parent.
    fingerprint_method : str
        Function to calculate fingerprint with. Accepts a SMILES, by default uses
        RDKFingerprints
    fingerprint_args : str
        Arguments for the fingerprint_method, by default None.
    similarity_method : str
        Function to calculate similarity with. Accepts two fingerprints
        and returns a similarity score.

    Attributes
    ----------
    crit_similarity : float
        The similarity similarity score threshold.
    increasing_similarity : bool
        Whether or not to only keep compounds whos similarity score is higher than its
        parent.
    fingerprint_method : str
        Fingerprint calculation method.
    fingerprint_args : dict
        Arguments for fingerprint method.
    similairty_method : str
        Simlarity calulation method.
    target_fps : List
        List of fingerprints.
    """

    def __init__(
        self,
        crit_similarity: float,
        increasing_similarity: bool,
        fingerprint_method: str = "RDKit",
        fingerprint_args: dict = None,
        similarity_method: str = "Tanimoto",
    ) -> None:
        self._filter_name = "Similarity Cutoff"
        self.crit_similarity = crit_similarity
        self.increasing_similarity = increasing_similarity

        self.fingerprint_method = fingerprint_method
        self.fingerprint_args = fingerprint_args or dict()
        self.similarity_method = similarity_method

        self.target_fps = []

    @property
    def filter_name(self) -> str:
        return self._filter_name

    def _set_target_fps(self, pickaxe: Pickaxe):
        self.target_fps = []
        for smiles in pickaxe.target_smiles:
            mol = MolFromSmiles(smiles)
            if self.fingerprint_method == "Morgan":
                fingerprint_args = self.fingerprint_args["radius"]
                fp = AllChem.GetMorganFingerprintAsBitVect(mol, fingerprint_args)
            else:
                fp = RDKFingerprint(mol)

            self.target_fps.append(fp)

    def _choose_items_to_filter(self, pickaxe: Pickaxe, processes: int = 1) -> Set[str]:
        """
        Compares the current generation to the target compound fingerprints
        marking compounds, who have a similarity score to a target
        compound greater than or equal to the crit_similarity, for expansion.
        """
        if not self.target_fps:
            self._set_target_fps(pickaxe)
        if not self.target_fps:
            print("No targets to filter for. Can't expand.")
            return [], []

        # Set up variables required for filtering
        # similarity Threshold
        if type(self.crit_similarity) in [list, tuple]:
            if len(self.crit_similarity) - 1 < pickaxe.generation:
                crit_similarity = self.crit_similarity[-1]
            else:
                crit_similarity = self.crit_similarity[pickaxe.generation]
        else:
            crit_similarity = self.crit_similarity

        # Get compounds eligible for expansion in the current generation
        compounds_to_check = []

        for cpd in pickaxe.compounds.values():
            # Compounds are in generation and correct type
            if cpd["Generation"] == pickaxe.generation and cpd["Type"] not in [
                "Coreactant",
                "Target Compound",
            ]:

                # Check for targets and only react if terminal
                if pickaxe.react_targets:
                    compounds_to_check.append(cpd)
                else:
                    # TODO this is not efficient
                    for t_id in pickaxe.targets:
                        if "C" + t_id[1:] != cpd["_id"]:
                            compounds_to_check.append(cpd)
                        else:
                            pickaxe.compounds[cpd["_id"]]["Expand"] = False

        # Run the filtering code to get a list of compounds to ignore
        print(
            f"Filtering Generation {pickaxe.generation} "
            f"with similarity > {crit_similarity}."
        )
        # Get input to filter code, c_id and smiles (to be
        # turned into fingerprint)
        cpd_info = [(cpd["_id"], cpd["SMILES"]) for cpd in compounds_to_check]
        if type(self.crit_similarity) in [list, tuple]:
            if len(self.crit_similarity) - 1 < pickaxe.generation:
                this_gen_crit_similarity = self.crit_similarity[-1]
            else:
                this_gen_crit_similarity = self.crit_similarity[pickaxe.generation]
        else:
            this_gen_crit_similarity = self.crit_similarity
        cpd_filters = self._filter_by_similarity_helper(
            cpd_info, self.target_fps, processes, this_gen_crit_similarity
        )

        # Process filtering results
        cpds_remove_set = set()
        for c_id, current_similarity in cpd_filters:
            # Check if similarity is increasing
            if self.increasing_similarity:
                if current_similarity >= pickaxe.compounds[c_id]["last_similarity"]:
                    pickaxe.compounds[c_id]["last_similarity"] = current_similarity
                else:
                    pickaxe.compounds[c_id]["Expand"] = False
                    cpds_remove_set.add(c_id)
                    continue

            if current_similarity < this_gen_crit_similarity:
                pickaxe.compounds[c_id]["Expand"] = False
                cpds_remove_set.add(c_id)

        return cpds_remove_set, []

    def _filter_by_similarity_helper(
        self,
        compounds_info: List[Tuple[str, str]],
        target_fps: List[AllChem.RDKFingerprint],
        processes: int,
        this_crit_similarity: float,
    ) -> List[Tuple[str, float]]:
        def print_progress(done: int, total: int, section: str) -> None:
            # Use print_on to print % completion roughly every 5 percent
            # Include max to print no more than once per compound (e.g. if
            # less than 20 compounds)
            print_on = max(round(0.1 * total), 1)
            if not (done % print_on):
                print(f"{section} {round(done / total * 100)} percent complete")

        # compound_info = [(smiles, id)]
        cpds_to_filter = list()
        compare_target_fps_partial = partial(
            self._compare_target_fps, target_fps, this_crit_similarity
        )

        if processes > 1:
            # Set up parallel computing of compounds to
            chunk_size = max([round(len(compounds_info) / (processes * 4)), 1])
            pool = multiprocessing.Pool(processes)
            for i, res in enumerate(
                pool.imap_unordered(
                    compare_target_fps_partial, compounds_info, chunk_size
                )
            ):
                # If the result of comparison is false, compound is not expanded
                # Default value for a compound is True, so no need to
                # specify expansion
                if res:
                    cpds_to_filter.append(res)
                print_progress(i, len(compounds_info), "Similarity filter progress:")

        else:
            for i, cpd in enumerate(compounds_info):
                res = compare_target_fps_partial(cpd)
                if res:
                    cpds_to_filter.append(res)
                print_progress(i, len(compounds_info), "Similarity filter progress:")
        print("Similarity filter progress: 100 percent complete")

        return cpds_to_filter

    def _compare_target_fps(
        self,
        target_fps: List[AllChem.RDKFingerprint],
        this_crit_similarity: float,
        compound_info: Tuple[str, str],
    ) -> Tuple[str, float]:
        # do finger print loop here
        """
        Helper function to allow parallel computation of Similarity filtering.
        Works with _filter_by_similarity_helper.

        Returns cpd_id if a the compound is similar enough to a target.
        """
        # Generate the fingerprint of a compound and compare to the fingerprints
        # of the targets
        def fingerprint(fingerprint_method, keyword_dict, smi):
            mol = AllChem.MolFromSmiles(smi)
            if fingerprint_method == "Morgan":
                return AllChem.GetMorganFingerprintAsBitVect(mol, **keyword_dict)
            else:
                return RDKFingerprint(mol)

        def calc_similarity(similarity_method, fp1, fp2):
            if similarity_method == "dice":
                return FingerprintSimilarity(
                    fp1, fp2, metric=DataStructs.DiceSimilarity
                )
            else:
                return FingerprintSimilarity(fp1, fp2)

        fingerprint_partial = partial(
            fingerprint, self.fingerprint_method, self.fingerprint_args
        )

        try:
            fp1 = fingerprint_partial(compound_info[1])
            max_similarity = 0
            for fp2 in target_fps:
                similarity = calc_similarity(self.similarity_method, fp1, fp2)
                if similarity >= this_crit_similarity:
                    return (compound_info[0], similarity)
                elif similarity >= max_similarity:
                    max_similarity = similarity
            return (compound_info[0], max_similarity)
            # TODO what except to use here?
        except:  # noqa
            return (compound_info[0], -1)

    def preprint(self, pickaxe: Pickaxe) -> None:
        if type(self.crit_similarity) in [list, tuple]:
            print_similarity = self.crit_similarity[pickaxe.generation]
        else:
            print_similarity = self.crit_similarity
        print(
            f"Filtering out compounds with maximum Similarity match < {print_similarity}"
        )

    def _post_print(
        self, pickaxe: Pickaxe, n_total: int, n_filtered: int, time_sample: float
    ) -> None:
        if type(self.crit_similarity) in [list, tuple]:
            if len(self.crit_similarity) - 1 < pickaxe.generation:
                print_similarity = self.crit_similarity[-1]
            else:
                print_similarity = self.crit_similarity[pickaxe.generation]
        else:
            print_similarity = self.crit_similarity
        print(
            (
                f"{n_filtered} of {n_total} compounds selected after "
                f"Similarity filtering of generation {pickaxe.generation} "
                f"at cutoff of {print_similarity}. "
                f"--took {round(time.time() - time_sample, 2)}s.\n"
            )
        )


class MCSFilter(Filter):
    """A filter that uses MCS score to determine compounds to expand.

    MCSFilter applies a strict cutoff to to the MCS score of
    compounds to determine which compounds to expand.

    Parameters
    ----------
    crit_mcs: float
        The maximum common substructure similarity score threshold.

    Attributes
    ----------
    crit_mcs : float
        The maximum common substructure similarity score threshold.
    """

    def __init__(self, crit_mcs: float) -> None:
        self._filter_name = "MCS Cutoff"
        self.crit_mcs = crit_mcs

    @property
    def filter_name(self) -> str:
        return self._filter_name

    def _choose_items_to_filter(self, pickaxe: Pickaxe, processes: int = 1) -> Set[str]:
        """
        Compares the current generation to the target compound fingerprints
        marking compounds, who have a similarity score to a target
        compound greater than or equal to the crit_similarity, for expansion.
        """

        if not pickaxe.target_smiles:
            print("No targets to filter for. Can't expand.")
            return [], []

        # Set up variables required for filtering
        # MCS Threshold
        if type(self.crit_mcs) in [list, tuple]:
            if len(self.crit_mcs) - 1 < pickaxe.generation:
                crit_mcs = self.crit_mcs[-1]
            else:
                crit_mcs = self.crit_mcs[pickaxe.generation]
        else:
            crit_mcs = self.crit_mcs

        # Get compounds eligible for expansion in the current generation
        compounds_to_check = []

        for cpd in pickaxe.compounds.values():
            # Compounds are in generation and correct type
            if cpd["Generation"] == pickaxe.generation and cpd["Type"] not in [
                "Coreactant",
                "Target Compound",
            ]:

                # Check for targets and only react if terminal
                if pickaxe.react_targets:
                    compounds_to_check.append(cpd)
                else:
                    for t_id in pickaxe.targets:
                        if "C" + t_id[1:] != cpd["_id"]:
                            compounds_to_check.append(cpd)
                        else:
                            pickaxe.compounds[cpd["_id"]]["Expand"] = False

        # Run the filtering code to get a list of compounds to ignore
        print(f"Filtering Generation {pickaxe.generation}" " with MCS > {crit_mcs}.")
        # Get input to filter code, c_id and smiles
        # (to be turned into fingerprint)
        cpd_info = [(cpd["_id"], cpd["SMILES"]) for cpd in compounds_to_check]
        this_gen_crit_mcs = crit_mcs
        cpd_filters = self._filter_by_mcs_helper(
            cpd_info, pickaxe.target_smiles, processes, this_gen_crit_mcs
        )

        # Process filtering results
        keep_ids = [cpd[0] for cpd in cpd_filters]
        cpds_remove_set = set()
        for c_id, _ in cpd_info:
            if c_id not in keep_ids:
                pickaxe.compounds[c_id]["Expand"] = False
                cpds_remove_set.add(c_id)

        return cpds_remove_set, []

    def _filter_by_mcs_helper(
        self,
        compounds_info: List[Tuple[str, str]],
        target_smiles: List[str],
        processes: int,
        this_crit_mcs: float,
        retro: bool = False,
    ) -> List[Tuple[str, float]]:
        def print_progress(done: int, total: int, section: str) -> None:
            # Use print_on to print % completion roughly every 5 percent
            # Include max to print no more than once per compound (e.g. if
            # less than 20 compounds)
            print_on = max(round(0.1 * total), 1)
            if not (done % print_on):
                print(f"{section} {round(done / total * 100)} percent complete")

        # compound_info = [(smiles, id)]
        cpds_to_filter = list()
        compare_target_mcs_partial = partial(
            self._compare_target_mcs, target_smiles, retro
        )

        if processes > 1:
            # Set up parallel computing of compounds to
            chunk_size = max([round(len(compounds_info) / (processes * 4)), 1])
            pool = multiprocessing.Pool(processes)
            for i, res in enumerate(
                pool.imap_unordered(
                    compare_target_mcs_partial,
                    compounds_info,
                    this_crit_mcs,
                    chunk_size,
                )
            ):

                if res:
                    cpds_to_filter.append(res)
                print_progress(
                    i,
                    len(compounds_info),
                    "Maximum Common Substructure filter progress:",
                )

        else:
            for i, cpd in enumerate(compounds_info):
                res = compare_target_mcs_partial(cpd, this_crit_mcs)
                if res:
                    cpds_to_filter.append(res)
                print_progress(
                    i,
                    len(compounds_info),
                    "Maximum Common Substructure filter progress:",
                )

        print("Maximum Common Substructure filter progress:" " 100 percent complete")
        return cpds_to_filter

    def _compare_target_mcs(
        self,
        target_smiles: List[str],
        retro: bool,
        compound_info: Tuple[str, str],
        this_crit_mcs: float,
    ) -> Tuple[str, float]:
        """Compare target MCS.

        Helper function to allow parallel computation of MCS filtering.
        Works with _filter_by_mcs_helper.

        Returns cpd_id if a the compound is similar enough to a target.

        """

        def get_mcs_overlap(mol, target_mol) -> float:
            mcs_out = mcs.FindMCS(
                [mol, target_mol], matchValences=False, ringMatchesRingOnly=False
            )

            if not mcs_out.canceled:
                ss_atoms = mcs_out.numAtoms
                ss_bonds = mcs_out.numBonds
                t_atoms = target_mol.GetNumAtoms()
                t_bonds = target_mol.GetNumBonds()

                mcs_overlap = (ss_atoms + ss_bonds) / (t_bonds + t_atoms)
                return mcs_overlap

            else:
                return 0

        # compare MCS for filter
        try:
            mol = AllChem.MolFromSmiles(compound_info[1])

            for t_smi in target_smiles:
                t_mol = AllChem.MolFromSmiles(t_smi)
                if not retro:
                    mcs_overlap = get_mcs_overlap(mol, t_mol)
                else:
                    mcs_overlap = get_mcs_overlap(t_mol, mol)

                if mcs_overlap > 1:
                    print("pause")
                if mcs_overlap >= this_crit_mcs:
                    return (compound_info[0], mcs_overlap)
        # TODO what except to use here?
        except:  # noqa
            return (compound_info[0], -1)

    def preprint(self, pickaxe: Pickaxe) -> None:
        if type(self.crit_mcs) in [list, tuple]:
            if len(self.crit_mcs) - 1 < pickaxe.generation:
                crit_mcs = self.crit_mcs[-1]
            else:
                crit_mcs = self.crit_mcs[pickaxe.generation]
        else:
            crit_mcs = self.crit_mcs
        print(f"Filtering out compounds with maximum MCS match < {crit_mcs}")

    def _post_print(
        self, pickaxe: Pickaxe, n_total: int, n_filtered: int, time_sample: float
    ) -> None:
        if type(self.crit_mcs) in [list, tuple]:
            if len(self.crit_mcs) - 1 < pickaxe.generation:
                crit_mcs = self.crit_mcs[-1]
            else:
                crit_mcs = self.crit_mcs[pickaxe.generation]
        else:
            crit_mcs = self.crit_mcs
        print(
            (
                f"{n_filtered} of {n_total} compounds selected after "
                f"MCS filtering of generation {pickaxe.generation} "
                f"at cutoff of {crit_mcs}. "
                f"--took {round(time.time() - time_sample, 2)}s.\n"
            )
        )


if __name__ == "__main__":
    pass



"""Pickaxe.py: Create network expansions from reaction rules and compounds.

This module generates new compounds from user-specified starting
compounds using a set of SMARTS-based reaction rules.
"""
import csv
import datetime
import os
import pickle
import time
from argparse import ArgumentParser
from io import StringIO
from pathlib import Path, PosixPath, WindowsPath
from sys import exit
from typing import List, Set, Tuple, Union

from rdkit.Chem import RemoveStereochemistry
from rdkit.Chem.AllChem import (
    AddHs,
    GetMolFrags,
    Kekulize,
    MolFromInchi,
    MolFromSmiles,
    MolToInchiKey,
    MolToSmiles,
    RDKFingerprint,
    ReactionFromSmarts,
    SanitizeMol,
)
from rdkit.Chem.Draw import MolToFile, rdMolDraw2D
from rdkit.Chem.rdchem import Mol
from rdkit.Chem.rdMolDescriptors import CalcMolFormula
from rdkit.RDLogger import logger

from minedatabase import utils
from minedatabase.databases import (
    MINE,
    write_compounds_to_mine,
    write_core_compounds,
    write_reactions_to_mine,
    write_targets_to_mine,
)
from minedatabase.reactions import transform_all_compounds_with_full


# Default to no errors
lg = logger()
lg.setLevel(4)


class Pickaxe:
    """Class to generate expansions with compounds and reaction rules.

    This class generates new compounds from user-specified starting
    compounds using a set of SMARTS-based reaction rules. It may be initialized
    with a text file containing the reaction rules and coreactants or this may
    be done on an ad hoc basis.

    Parameters
    ----------
    rule_list : str
        Filepath of rules.
    coreactant_list : str
        Filepath of coreactants.
    explicit_h : bool, optional
        Whether rules utilize explicit hydrogens, by default True.
    kekulize : bool, optional
        Whether or not to kekulize compounds before reaction,
        by default False.
    neutralise : bool, optional
        Whether or not to neutralise compounds, by default True.
    errors : bool, optional
        Whether or not to print errors to stdout, by default True.
    inchikey_blocks_for_cid : int, optional
        How many blocks of the InChI key to use for the compound id,
        by default 1.
    database : str, optional
        Name of the database where to save results, by default None.
    database_overwrite : bool, optional
        Whether or not to erase existing database in event of a collision,
        by default False.
    mongo_uri : bool, optional
        uri for the mongo client, by default 'mongodb://localhost:27017'.
    image_dir : str, optional
        Filepath where images should be saved, by default None.
    quiet : bool, optional
        Whether to silence warnings, by default False.
    react_targets : bool, optional
        Whether or not to apply reactions to generated compounds that match
        targets, by default True.
    filter_after_final_gen : bool, optional
        Whether to apply filters after final expansion, by default True.
    prune_between_gens : bool, optional
        Whether to prune network between generations if using filters

    Attributes
    ----------
    operators: dict
        Reaction operators to transform compounds with.
    coreactants: dict
        Coreactants required by the operators.
    compounds: dict
        Compounds in the pickaxe network.
    reactions: dict
        Reactions in the pickaxe network.
    generation: int
        The current generation
    explicit_h : bool
        Whether rules utilize explicit hydrogens.
    kekulize : bool
        Whether or not to kekulize compounds before reaction.
    neutralise : bool
        Whether or not to neutralise compounds.
    fragmented_mols : bool
        Whether or not to allow fragmented molecules.
    radical_check : bool
        Whether or not to check and remove radicals.
    image_dir : str, optional
        Filepath where images should be saved.
    errors : bool
        Whether or not to print errors to stdout.
    quiet : bool
        Whether or not to silence warnings.
    filters: List[object]
        A list of filters to apply during the expansion.
    targets : dict
        Molecules to be targeted during expansions.
    target_smiles: List[str]
        The SMILES of all the targets.
    react_targets : bool
        Whether or not to react targets when generated.
    filter_after_final_gen : bool
        Whether or not to filter after the last expansion.
    prune_between_gens : bool, optional
        Whether to prune network between generations if using filters.
    mongo_uri : str
        The connection string to the mongo database.
    cid_num_inchi_blocks : int
        How many blocks of the inchi-blocks to use to generate the compound id.
    """

    def __init__(
        self,
        rule_list: str = None,
        coreactant_list: str = None,
        explicit_h: bool = False,
        kekulize: bool = False,
        neutralise: bool = True,
        errors: bool = True,
        inchikey_blocks_for_cid: int = 1,
        database: str = None,
        database_overwrite: bool = False,
        mongo_uri: bool = "mongodb://localhost:27017",
        image_dir: str = None,
        quiet: bool = True,
        react_targets: bool = True,
        filter_after_final_gen: bool = True,
        prune_between_gens: bool = False,
    ) -> None:
        # Main pickaxe properties
        self.operators = {}
        self.coreactants = {}
        self.compounds = {}
        self.reactions = {}
        self.generation = 0
        # Chemistry options
        self.explicit_h = explicit_h
        self.kekulize = kekulize
        self.neutralise = neutralise
        self.fragmented_mols = False
        self.radical_check = False
        # Other options
        self.image_dir = image_dir
        self.errors = errors
        self.quiet = quiet
        # For filtering
        self.filters = []
        self.targets = dict()
        self.target_smiles = []
        self.react_targets = react_targets
        self.filter_after_final_gen = filter_after_final_gen
        self.prune_between_gens = prune_between_gens
        # database info
        self.mongo_uri = mongo_uri
        # partial_operators
        self.use_partial = False
        self.partial_operators = dict()
        # cid options
        self.cid_num_inchi_blocks = inchikey_blocks_for_cid

        print("----------------------------------------")
        print("Intializing pickaxe object")
        if database:
            # Determine if a specified database is legal
            db = MINE(database, self.mongo_uri)
            if database in db.client.list_database_names():
                if database_overwrite:
                    # If db exists, remove db from all of core compounds
                    # and drop db
                    print(
                        (
                            f"Database {database} already exists. "
                            "Deleting database and removing from core compound"
                            " mines."
                        )
                    )
                    db.core_compounds.update_many({}, {"$pull": {"MINES": database}})
                    db.client.drop_database(database)
                    self.mine = database
                else:
                    print(
                        (
                            f"Warning! Database {database} already exists."
                            "Specify database_overwrite as true to delete "
                            "old database and write new."
                        )
                    )
                    exit("Exiting due to database name collision.")
                    self.mine = None
            else:
                self.mine = database
            del db
        else:
            self.mine = None

        # Use RDLogger to catch errors in log file. SetLevel indicates mode (
        # 0 - debug, 1 - info, 2 - warning, 3 - critical). Default is no errors
        if errors:
            lg.setLevel(0)

        # Load coreactants (if any) into Pickaxe object
        if coreactant_list:
            with open(coreactant_list) as infile:
                for coreactant in infile:
                    self._load_coreactant(coreactant)

        # Load rules (if any) into Pickaxe object
        if rule_list:
            self._load_operators(rule_list)

        print("\nDone intializing pickaxe object")
        print("----------------------------------------\n")

    def load_targets(
        self,
        target_compound_file: Union[str, None],
        id_field: str = "id",
    ) -> None:
        """Load targets into pickaxe.

        Parameters
        ----------
        target_compound_file : str
            Filepath of target compounds.
        id_field : str, optional
            Header value of compound id in input file, by default 'id'.
        """
        if not target_compound_file:
            print("No target file given")
            return None

        for target_dict in utils.file_to_dict_list(target_compound_file):
            mol = self._mol_from_dict(target_dict)
            if not mol:
                continue
            # Add compound to internal dictionary as a target
            # compound and store SMILES string to be returned
            smi = MolToSmiles(mol, True)
            cpd_name = target_dict[id_field]
            # Only operate on organic compounds
            if "c" in smi.lower():
                SanitizeMol(mol)
                self._add_compound(cpd_name, smi, "Target Compound", mol)
                self.target_smiles.append(smi)

        print(f"{len(self.target_smiles)} target compounds loaded\n")

    def load_compound_set(self, compound_file: str = None, id_field: str = "id") -> str:
        """Load compounds for expansion into pickaxe.

        Parameters
        ----------
        compound_file : str, optional
            Filepath of compounds, by default None.
        id_field : str, optional
            Header value of compound id in input file, by default 'id'.

        Returns
        -------
        str
            List of SMILES that were succesfully loaded into pickaxe.

        Raises
        ------
        ValueError
            No file specified for loading.
        """

        # load compounds
        initial_cpds_n = len(self.compounds)
        compound_smiles = []
        if compound_file:
            for cpd_dict in utils.file_to_dict_list(compound_file):
                mol = self._mol_from_dict(cpd_dict)
                if not mol:
                    continue
                RemoveStereochemistry(mol)
                # Add compound to internal dictionary as a starting
                # compound and store SMILES string to be returned
                smi = MolToSmiles(mol, True)
                cpd_name = cpd_dict[id_field]
                # Do not operate on inorganic compounds
                if "C" in smi or "c" in smi:
                    # resolve potential tautomers and choose first one
                    if "n" in smi:
                        smi = utils.postsanitize_smiles([smi])[0][0]
                        mol = MolFromSmiles(smi)
                    if mol:
                        SanitizeMol(mol)
                        self._add_compound(
                            cpd_name, smi, cpd_type="Starting Compound", mol=mol
                        )
                        compound_smiles.append(smi)
                    else:
                        print("Warning: could not load compound:", smi)

        else:
            raise ValueError("No input file specified for starting compounds")

        print(f"{len(compound_smiles)} compounds loaded...")
        print(
            f"({len(self.compounds) - initial_cpds_n} "
            f"after removing stereochemistry)"
        )

        return compound_smiles

    def _load_coreactant(self, coreactant_text: str) -> None:
        """load_coreactants into pickaxe.

        Parameters
        ----------
        coreactant_text : str
            A line read in from the coreactant data file.
        """

        # If coreactant is commented out (with '#') then don't import
        if coreactant_text[0] == "#":
            return
        split_text = coreactant_text.strip().split("\t")
        # split_text[0] is compound name, split_text[1] is SMILES string
        # Generate a Mol object from the SMILES string if possible
        try:
            mol = MolFromSmiles(split_text[2])
            if not mol:
                raise ValueError
            # TODO: what do do about stereochemistry? Original comment is below
            # but stereochem was taken out (isn't it removed later anyway?)
            # # Generate SMILES string with stereochemistry taken into account
            smi = MolToSmiles(mol)
        except (IndexError, ValueError):
            raise ValueError(f"Unable to load coreactant: {coreactant_text}")
        cpd_id = self._add_compound(split_text[0], smi, "Coreactant", mol)
        # If hydrogens are to be explicitly represented, add them to the Mol
        # object
        if self.explicit_h:
            mol = AddHs(mol)
        # If kekulization is preferred (no aromatic bonds, just 3 C=C bonds
        # in a 6-membered aromatic ring for example)
        if self.kekulize:
            Kekulize(mol, clearAromaticFlags=True)
        # Store coreactant in a coreactants dictionary with the Mol object
        # and hashed id as values (coreactant name as key)
        self.coreactants[split_text[0]] = (
            mol,
            cpd_id,
        )

    def _load_operators(self, rule_path: str) -> None:
        """Load reaction rules into pickaxe.

        Parameters
        ----------
        rule_path : str
            Filepath of reaction rules.
        """
        skipped = 0

        # Get the stream for rule input
        if type(rule_path) in [str, Path, PosixPath, WindowsPath]:
            infile = open(rule_path)
        elif type(rule_path) == StringIO:
            infile = rule_path

        with infile:
            # Get all reaction rules from tsv file and store in dict (rdr)
            rdr = csv.DictReader(
                (row for row in infile if not row.startswith("#")), delimiter="\t"
            )
            for rule in rdr:
                try:
                    # Get reactants and products for each reaction into list
                    # form (not ; delimited string)
                    rule["Reactants"] = rule["Reactants"].split(";")
                    rule["Products"] = rule["Products"].split(";")
                    # Ensure that all coreactants are known and accounted for
                    all_rules = rule["Reactants"] + rule["Products"]
                    for coreactant_name in all_rules:
                        if (
                            coreactant_name not in self.coreactants
                            and coreactant_name != "Any"
                        ):
                            raise ValueError(
                                "Undefined coreactant:" f"{coreactant_name}"
                            )
                    # Create ChemicalReaction object from SMARTS string
                    rxn = ReactionFromSmarts(rule["SMARTS"])
                    rule.update(
                        {
                            "_id": rule["Name"],
                            "Reactions_predicted": 0,
                            "SMARTS": rule["SMARTS"],
                        }
                    )
                    # Ensure that we have number of expected reactants for
                    # each rule
                    if rxn.GetNumReactantTemplates() != len(
                        rule["Reactants"]
                    ) or rxn.GetNumProductTemplates() != len(rule["Products"]):
                        skipped += 1
                        print(
                            "The number of coreactants does not match the "
                            "number of compounds in the SMARTS for reaction "
                            "rule: " + rule["Name"]
                        )
                    if rule["Name"] in self.operators:
                        raise ValueError("Duplicate reaction rule name")
                    # Update reaction rules dictionary
                    self.operators[rule["Name"]] = (rxn, rule)
                except Exception as e:
                    raise ValueError(f"{str(e)}\nFailed to parse" f"{rule['Name']}")
        if skipped:
            print(f"WARNING: {skipped} rules skipped")

    def _mol_from_dict(self, input_dict: dict) -> Mol:
        """Generate an RDKit mol object from a dictionary.

        Parameters
        ----------
        input_dict : dict
            Should have "smiles", "inchi", and "structure" fields.

        Returns
        -------
        mol : Mol
            Mol object created from structure string in input_dict.

        Raises
        ------
        ValueError
            Structure field not found in input.
        """
        structure_field = None
        for field in input_dict:
            if str(field).lower() in {"smiles", "inchi", "structure"}:
                structure_field = field
                break

        if not structure_field:
            raise ValueError("Structure field not found in input.")

        if structure_field not in input_dict:
            return
        # Generate Mol object from InChI code if present
        if "InChI=" in input_dict[structure_field]:
            mol = MolFromInchi(input_dict[structure_field])
        # Otherwise generate Mol object from SMILES string
        else:
            mol = MolFromSmiles(input_dict[structure_field])
        if not mol:
            if self.errors:
                print(f"Unable to Parse {input_dict[structure_field]}")
            return
        # If compound is disconnected (determined by GetMolFrags
        # from rdkit) and loading of these molecules is not
        # allowed (i.e. fragmented_mols == 1), then don't add to
        # internal dictionary. This is most common when compounds
        # are salts.
        if not self.fragmented_mols and len(GetMolFrags(mol)) > 1:
            return
        # If specified remove charges (before applying reaction
        # rules later on)
        if self.neutralise:
            mol = utils.neutralise_charges(mol)
        return mol

    def _gen_compound(
        self, cpd_name: str, smi: str, cpd_type: str, mol: Mol = None
    ) -> Tuple[str, dict]:
        """Generate a compound.

        Parameters
        ----------
        cpd_name : str
            Name of the compound to add.
        smi : str
            SMILES of the compound to add.
        cpd_type : str
            Type of compound
        mol : Mol, optional
            RDKit Molecule, by default None.

        Returns
        -------
        Tuple[str, dict]
            Compound id and compound dict.
        """
        # resolve potential tautomers and choose first one
        try:
            if "n" in smi:
                smi = utils.postsanitize_smiles([smi])[0][0]

            cpd_dict = {}
            cpd_id, inchi_key = utils.get_compound_hash(
                smi, cpd_type, self.cid_num_inchi_blocks
            )
        except BaseException:
            return None, None

        if cpd_id:
            # We don't want to overwrite the same compound from a prior
            # generation so we check with hashed id from above
            if cpd_id not in self.compounds:
                if not mol:
                    mol = MolFromSmiles(smi)
                # expand only Predicted and Starting_compounds
                expand = cpd_type in ["Predicted", "Starting Compound"]
                cpd_dict = {
                    "ID": cpd_name,
                    "_id": cpd_id,
                    "SMILES": smi,
                    "InChI_key": inchi_key,
                    "Type": cpd_type,
                    "Generation": self.generation,
                    "atom_count": utils.get_atom_count(mol, self.radical_check),
                    "Reactant_in": [],
                    "Product_of": [],
                    "Expand": expand,
                    "Formula": CalcMolFormula(mol),
                    "last_tani": 0,
                }
                if cpd_id.startswith("X"):
                    del cpd_dict["Reactant_in"]
                    del cpd_dict["Product_of"]
            else:
                cpd_dict = self.compounds[cpd_id]

            return cpd_id, cpd_dict
        else:
            return None, None

    def _add_compound(
        self, cpd_name: str, smi: str, cpd_type: str, mol: Mol = None
    ) -> str:
        """Generate a compound.

        Parameters
        ----------
        cpd_name : str
            Name of the compound to add.
        smi : str
            SMILES of the compound to add.
        cpd_type : str
            Type of compound.
        mol : Mol, optional
            RDKit Molecule, by default None.

        Returns
        -------
        cpd_id : str
            Compound ID.
        """

        # We don't want to overwrite the same compound from a prior
        # generation so we check with hashed id from above
        cpd_id, cpd_dict = self._gen_compound(cpd_name, smi, cpd_type, mol)
        if cpd_id:
            if (cpd_type == "Target Compound") and (cpd_id not in self.targets):
                self.targets[cpd_id] = cpd_dict
            elif cpd_id not in self.compounds:
                self.compounds[cpd_id] = cpd_dict

                if self.image_dir and self.mine:
                    try:
                        with open(
                            os.path.join(self.image_dir, cpd_id + ".svg"), "w"
                        ) as outfile:

                            mol = MolFromSmiles(cpd_dict["SMILES"])
                            nmol = rdMolDraw2D.PrepareMolForDrawing(mol)
                            d2d = rdMolDraw2D.MolDraw2DSVG(1000, 1000)
                            d2d.DrawMolecule(nmol)
                            d2d.FinishDrawing()
                            outfile.write(d2d.GetDrawingText())
                    except OSError:
                        print(f"Unable to generate image for {cpd_dict['SMILES']}")

        return cpd_id

    def transform_all(self, processes: int = 1, generations: int = 1) -> None:
        """Transform compounds with reaction operators.

        Apply reaction rules to compounds and generate a specified number
        of new generations.

        Parameters
        ----------
        processes : int, optional
            Number of processes to run in parallel, by default 1.
        generations : int, optional
            Number of generations to create, by default 1.
        """

        while self.generation < generations or (
            self.generation == generations and self.filter_after_final_gen
        ):

            for _filter in self.filters:
                _filter.apply_filter(self, processes, generation=generations)

            if self.generation < generations:
                # Prune network to only things that are expanded as white list
                if self.prune_between_gens and self.filters:
                    # Find targets and compounds where expand is true
                    white_list = []
                    for cpd_id, cpd_info in self.compounds.items():
                        if cpd_id.startswith("T"):
                            continue
                        elif cpd_info["Expand"] or cpd_id.startswith("X"):
                            white_list.append(cpd_id)
                        elif f"T{cpd_id[1:]}" in self.targets:
                            white_list.append(cpd_id)

                    # Prune network to only things being expanded
                    self.prune_network(white_list, True)

                print("----------------------------------------")
                print(f"Expanding Generation {self.generation + 1}\n")

                # Starting time for expansion
                time_init = time.time()

                # Tracking compounds formed
                n_comps = len(self.compounds)
                n_rxns = len(self.reactions)

                # Get SMILES to be expanded
                compound_smiles = [
                    cpd["SMILES"]
                    for cpd in self.compounds.values()
                    if cpd["Generation"] == self.generation
                    and cpd["Type"] not in ["Coreactant", "Target Compound"]
                    and cpd["Expand"]
                ]
                # No compounds found
                if not compound_smiles:
                    print(
                        "No compounds to expand in generation "
                        f"{self.generation + 1}. Finished expanding."
                    )
                    return None

                self._transform_helper(compound_smiles, processes)
                self._remove_cofactor_redundancy()

                print(
                    f"Generation {self.generation + 1} finished in"
                    f" {time.time()-time_init} s and contains:"
                )
                print(f"\t\t{len(self.compounds) - n_comps} new compounds")
                print(f"\t\t{len(self.reactions) - n_rxns} new reactions")
                print(f"\nDone expanding Generation: {self.generation + 1}.")
                print("----------------------------------------\n")

            self.generation += 1

    # Partial operator code
    # def load_partial_operators(self, mapped_reactions):
    #     """Generate set of partial operators from a list of mapped reactions
    #     corresponding to the reaction rules being used.
    #     :param mapped_reactions: A .csv file with four columns: rule id,
    #     source, SMARTS, mapping info.
    #     :type mapped_reactions: file
    #     """
    #     # generate partial operators as done in ipynb
    #     if not self.operators:
    #         print("Load reaction rules before loading partial operators")
    #     else:
    #         with open(mapped_reactions) as f:
    #             for line in f.readlines():
    #                 # Grab info from current mapped reaction
    #                 rule, source, smiles, _ = line.strip('\n').split('\t')
    #                 # There should be 2 or more reactants derived from
    #                 # the mapping code The mapped code doesn't include
    #                 # cofactors, so 2 or more means any;any*
    #                 exact_reactants = smiles.split('>>')[0]\
    #                                         .replace(';', '.').split('.')

    #                 base_rule = rule.split('_')[0]
    #                 # base rule must be loaded for partial operator
    #                 # to be used
    #                 if base_rule in self.operators:
    #                     op_reactants = (
    #                       self.operators[base_rule][1]['Reactants']
    #                     )
    #                     if op_reactants.count('Any') >= 2:
    #                         mapped_reactants = []
    #                         for i, r in enumerate(op_reactants):
    #                             if r == 'Any':
    #                                 mapped_reactants.append(
    #                                     exact_reactants.pop(0)
    #                                     )
    #                             else:
    #                                 mapped_reactants.append(r)

    #                         ind_SMARTS = (
    #                           self.operators[base_rule][1]['SMARTS']
    #                         )
    #                         ind_SMARTS = (ind_SMARTS.split('>>')[0].
    #                                       split('>>')[0].replace('(', '').
    #                                       replace(')', '').split('.'))
    #                         # now loop through and generate dictionary entries
    #                         for i, r in enumerate(op_reactants):
    #                             if r != 'Any':
    #                                 pass
    #                             else:
    #                                 # Build entries
    #                                 fixed_reactants = [
    #                                     fr if i != j else 'SMARTS_match'
    #                                     for j, fr in
    #                                     enumerate(mapped_reactants)
    #                                 ]

    # bi_rule =  {
    #     'rule': base_rule,
    #     'rule_reaction': rule,
    #     'reactants': fixed_reactants
    # }
    # if (ind_SMARTS[i] in self.partial_operators:
    #     self.partial_operators[ind_SMARTS[i]].append(bi_rule)
    # else:
    #     self.partial_operators[ind_SMARTS[i]] = [bi_rule]

    # def _filter_partial_operators(self):
    #     # generate the reactions to specifically expand
    #     # based on current compounds
    #     def partial_reactants_exist(partial_rule):
    #         try:
    #             rule_reactants = (
    #               self.operators[partial_rule['rule']][1]['Reactants']
    #             )
    #             cofactor = [False if r == 'Any' else True
    #                         for r in rule_reactants]

    #             reactant_ids = []
    #         for is_cofactor, smi in zip(cofactor, partial_rule['reactants']):
    #                 if is_cofactor:
    #                     reactant_ids.append(self.coreactants[smi][1])
    #                 elif smi == 'SMARTS_match':
    #                     continue
    #                 else:
    #                     reactant_ids.append(utils.get_compound_hash(smi))

    #             reactants_exist = [r in self.compounds for r in reactant_ids]
    #             if all(reactants_exist):
    #                 return True
    #             else:
    #                 return False
    #         except:
    #             return False

    #     filtered_partials = dict()
    #     for SMARTS_match, rules in self.partial_operators.items():
    #         for rule in rules:
    #             if partial_reactants_exist(rule):
    #                 if SMARTS_match in filtered_partials:
    #                     filtered_partials[SMARTS_match].append(rule)
    #                 else:
    #                     filtered_partials[SMARTS_match] = [rule]

    #     return filtered_partials

    def _remove_cofactor_redundancy(self) -> None:
        """Check for reactions that are to be removed.

        Checks for cofactors in rxns that were generated by an any;any rules
        and are specified as generated compounds. Removes redundant reactions
        and ensures cofactors are all labeled as cofactors.
        """
        # Identify compounds who are really cofactors
        cofactors_as_cpds = []
        cofactor_ids = [cofactor[1] for cofactor in self.coreactants.values()]
        for cpd_id in self.compounds:
            if "X" + cpd_id[1:] in cofactor_ids and cpd_id.startswith("C"):
                cofactors_as_cpds.append(cpd_id)

        # Loop through identified compounds and update
        # reactions/compounds accordingly
        rxns_to_del = set()
        for cpd_id in cofactors_as_cpds:
            rxn_ids = set(
                self.compounds[cpd_id]["Product_of"]
                + self.compounds[cpd_id]["Reactant_in"]
            )
            rxns_to_del = rxns_to_del.union(rxn_ids)
            # Check and fix reactions as needed
            for rxn_id in rxn_ids:
                rxn = self.reactions[rxn_id]
                # generate products list with replacements
                reactants = []
                products = []
                for s, reactant in rxn["Reactants"]:
                    if reactant in cofactors_as_cpds:
                        reactants.append((s, self.compounds["X" + reactant[1:]]))
                    else:
                        reactants.append((s, self.compounds[reactant]))

                for s, product in rxn["Products"]:
                    if product in cofactors_as_cpds:
                        products.append((s, self.compounds["X" + product[1:]]))
                    else:
                        products.append((s, self.compounds[product]))

                cofactor_rxn_id, rxn_text = utils.get_reaction_hash(reactants, products)

                # Check to see if reaction makes changes
                # Reactions such as
                # '(1) NCCCNc1cc(C(=O)O)ccc1O + (1) O =>
                #       (1) O + (1) NCCCNc1cc(C(=O)O)ccc1O'
                # are possible. Filter out
                sorted_reactants = sorted(reactants, key=lambda x: x[1]["_id"])
                sorted_products = sorted(products, key=lambda x: x[1]["_id"])

                # Update newly calculated reaction info in self.reactions.
                # Four possibilites:
                # 1. Reaction has same products as reactants -- Remove all
                # 2. Cofactors only as reactants
                # 3. Reaction exists already -- append to existing reaction
                # 4. Reaction doesn't exist -- make new reaction

                # After this, delete redudnant compounds and update
                # other compounds in reaction

                append_new = False
                # Reaction is something like CoF + CoF -> Product + Product
                # Where products are actually just cofactors
                if sorted_reactants == sorted_products:
                    pass  # No need to update anything in self.reactions

                # All reactants are cofactors, don't create reaction
                elif all([r[1]["_id"].startswith("X") for r in sorted_reactants]):
                    pass

                # Reaction already exists, update reaction operators
                elif cofactor_rxn_id in self.reactions:
                    ops = self.reactions[cofactor_rxn_id]["Operators"]
                    ops = ops | self.reactions[rxn_id]["Operators"]
                    self.reactions[cofactor_rxn_id]["Operators"] = ops
                    append_new = True

                # Reaction does not exist, generate new reaction
                else:
                    cofactor_rxn = {
                        "_id": cofactor_rxn_id,
                        # give stoich and id of reactants/products
                        "Reactants": [(s, r["_id"]) for s, r in reactants],
                        "Products": [(s, p["_id"]) for s, p in products],
                        "Operators": rxn["Operators"],
                        "SMILES_rxn": rxn_text,
                    }

                    # Assign reaction to reactions dict
                    self.reactions[cofactor_rxn_id] = cofactor_rxn
                    append_new = True

                for _, cpd in rxn["Reactants"]:
                    if cpd.startswith("C"):
                        if rxn_id in self.compounds[cpd]["Reactant_in"]:
                            self.compounds[cpd]["Reactant_in"].remove(rxn_id)

                        if (
                            append_new
                            and cofactor_rxn_id
                            not in self.compounds[cpd]["Reactant_in"]
                        ):
                            self.compounds[cpd]["Reactant_in"].append(cofactor_rxn_id)

                for _, cpd in rxn["Products"]:
                    if cpd.startswith("C"):
                        if rxn_id in self.compounds[cpd]["Product_of"]:
                            self.compounds[cpd]["Product_of"].remove(rxn_id)

                        if append_new and (
                            cofactor_rxn_id not in self.compounds[cpd]["Product_of"]
                        ):
                            self.compounds[cpd]["Product_of"].append(cofactor_rxn_id)

        if rxns_to_del:
            for rxn_id in rxns_to_del:
                # Remove any links from products/reactants
                # TODO: this should have happened in above loop
                # but getting errors stemming from this
                for _, cpd in self.reactions[rxn_id]["Reactants"]:
                    if cpd.startswith("C"):
                        if rxn_id in self.compounds[cpd]["Reactant_in"]:
                            self.compounds[cpd]["Reactant_in"].remove(rxn_id)

                for _, cpd in self.reactions[rxn_id]["Products"]:
                    if cpd.startswith("C"):
                        if rxn_id in self.compounds[cpd]["Product_of"]:
                            self.compounds[cpd]["Product_of"].remove(rxn_id)

                del self.reactions[rxn_id]

            for cpd_id in cofactors_as_cpds:
                del self.compounds[cpd_id]

        orphan_cpds = []
        # Check for orphaned compounds
        for cpd_id, cpd in self.compounds.items():
            if cpd_id.startswith("C"):
                if (
                    (not cpd["Reactant_in"])
                    and (not cpd["Product_of"])
                    and (cpd["Type"] != "Starting Compound")
                ):
                    orphan_cpds.append(cpd_id)

        for cpd_id in orphan_cpds:
            del self.compounds[cpd_id]

    def prune_network(self, white_list: list, print_output: str = True) -> None:
        """Prune the reaction network to a list of targets.

        Prune the predicted reaction network to only compounds and reactions
        that terminate in a specified white list of compounds.

        Parameters
        ----------
        white_list : list
            A list of compound ids to filter the network to.
        print_output : bool
            Whether or not to print output
        """
        n_white = len(white_list)
        cpd_set, rxn_set = self.find_minimal_set(white_list)
        self.compounds = dict(
            [(k, v) for k, v in self.compounds.items() if k in cpd_set]
        )
        self.reactions = dict(
            [(k, v) for k, v in self.reactions.items() if k in rxn_set]
        )

        if print_output:
            print(
                f"Pruned network to {len(cpd_set)} compounds and "
                f"{len(rxn_set)} reactions based on "
                f"{n_white} whitelisted compounds.\n"
            )

    def prune_network_to_targets(self) -> None:
        """Prune the reaction network to the target compounds.

        Prune the predicted reaction network to only compounds and reactions
        that terminate in the target compounds.
        """
        print("----------------------------------------")
        prune_start = time.time()
        white_list = set()
        for target_id in self.targets:
            white_list.add("C" + target_id[1:])

        print(f"Pruning to {len(white_list)} target compounds")
        self.prune_network(white_list)

        n_targets = 0
        for cpd_id in self.compounds:
            if f"T{cpd_id[1:]}" in self.targets:
                n_targets += 1

        print(f"Found {n_targets} targets.")
        print(f"Pruning took {time.time() - prune_start}s")
        print("----------------------------------------\n")

    def find_minimal_set(self, white_list: Set[str]) -> Tuple[set, set]:
        """Find the minimal set of compounds and reactions given a white list.

        Given a whitelist this function finds the minimal set of compound and
        reactions ids that comprise the set.

        Parameters
        ----------
        white_list : Set[str]
            List of compound_ids to use to filter reaction network to.

        Returns
        -------
        Tuple[set, set]
            The filtered compounds and reactions.
        """

        queue = list(white_list)
        visited = set()
        cpd_set = set()
        rxn_set = set()

        while queue:
            cpd_id = queue.pop()
            visited.add(cpd_id)
            # Select compound from whitelist
            if cpd_id not in self.compounds:
                continue
            else:
                cpd_set.add(cpd_id)

            # Add info for reactions that produce compound
            for rxn_id in self.compounds[cpd_id].get("Product_of", []):
                rxn_set.add(rxn_id)
                # Add products, not to be further explored
                for cpd in self.reactions[rxn_id]["Products"]:
                    cpd_set.add(cpd[1])

                # Add reactants, also add to queue
                for cpd in self.reactions[rxn_id]["Reactants"]:
                    cpd_set.add(cpd[1])
                    # Termination conditions
                    if (
                        cpd[1].startswith("C")
                        and cpd[1] not in visited
                        and self.compounds[cpd[1]]["Type"] != "Starting Compound"
                    ):
                        queue.append(cpd[1])

        return cpd_set, rxn_set

    def assign_ids(self) -> None:
        """Assign a numerical ID to compounds (and reactions).

        Assign IDs that are unique only to the CURRENT run.
        """
        # TODO is this necessary?
        # # If we were running a multiprocess expansion, this removes the dicts
        # # from Manager control
        # self.compounds = dict(self.compounds)
        # self.reactions = dict(self.reactions)
        i = 1
        for comp in sorted(
            self.compounds.values(), key=lambda x: (x["Generation"], x["_id"])
        ):
            # Create ID of form ####### ending with i, padded with zeroes to
            # fill unused spots to the left with zfill (e.g. ID = '0003721' if
            # i = 3721).
            if not comp.get("ID"):
                comp["ID"] = "pkc" + str(i).zfill(7)
                i += 1
                self.compounds[comp["_id"]] = comp
                # If we are not loading into the mine, we generate the image
                # here.
                if self.image_dir and not self.mine:
                    mol = MolFromSmiles(comp["SMILES"])
                    try:
                        MolToFile(
                            mol,
                            os.path.join(self.image_dir, comp["ID"] + ".png"),
                            fitImage=True,
                            kekulize=False,
                        )
                    except OSError:
                        print(f"Unable to generate image for {comp['SMILES']}")
        i = 1
        for rxn in sorted(self.reactions.values(), key=lambda x: x["_id"]):
            rxn["ID_rxn"] = (
                " + ".join(
                    [
                        f"({x[0]}) {self.compounds[x[1]]['ID']}[c0]"
                        for x in rxn["Reactants"]
                    ]
                )
                + " => "
                + " + ".join(
                    [
                        f"({x[0]}) {self.compounds[x[1]]['ID']}[c0]"
                        for x in rxn["Products"]
                    ]
                )
            )
            # Create ID of form ####### ending with i, padded with zeroes to
            # fill unused spots to the left with zfill (e.g. ID = '0003721' if
            # i = 3721).
            rxn["ID"] = "pkr" + str(i).zfill(7)
            i += 1
            self.reactions[rxn["_id"]] = rxn

    def write_compound_output_file(self, path: str, dialect: str = "excel-tab") -> None:
        """Write compounds to an output file.

        Parameters
        ----------
        path : str
            Path to write data.
        dialect : str, optional
            Dialect of the output, by default 'excel-tab'.
        """
        path = utils.prevent_overwrite(path)

        columns = ("ID", "Type", "Generation", "Formula", "InChIKey", "SMILES")
        for _id, val in self.compounds.items():
            inchi_key = MolToInchiKey(MolFromSmiles(val["SMILES"]))
            self.compounds[_id]["InChIKey"] = inchi_key

        with open(path, "w") as outfile:
            writer = csv.DictWriter(
                outfile,
                columns,
                dialect=dialect,
                extrasaction="ignore",
                lineterminator="\n",
            )
            writer.writeheader()
            writer.writerows(sorted(self.compounds.values(), key=lambda x: x["ID"]))

    def write_reaction_output_file(self, path: str, delimiter: str = "\t") -> None:
        r"""Write all reaction data to the specified path.

        Parameters
        ----------
        path : str
            Path to write data.
        delimiter : str, optional
            Delimiter for the output file, by default '\t'.
        """
        path = utils.prevent_overwrite(path)
        with open(path, "w") as outfile:
            outfile.write(
                "ID\tName\tID equation\tSMILES equation\tRxn hash\t" "Reaction rules\n"
            )
            for rxn in sorted(self.reactions.values(), key=lambda x: x["ID"]):
                outfile.write(
                    delimiter.join(
                        [
                            rxn["ID"],
                            "",
                            rxn["ID_rxn"],
                            rxn["SMILES_rxn"],
                            rxn["_id"],
                            ";".join(rxn["Operators"]),
                        ]
                    )
                    + "\n"
                )

    def save_to_mine(
        self, processes: int = 1, indexing: bool = True, write_core: bool = False
    ) -> None:
        """Save pickaxe run to MINE database.

        Parameters
        ----------
        processes : int, optional
            Number of processes to use, by default 1.
        indexing : bool, optional
            Whether or not to add indexes, by default True.
        write_core : bool, optional
            Whether or not to write to core database, by default False.
        """
        print("\n----------------------------------------")
        print(f"Writing results to {self.mine} Database")
        print("----------------------------------------\n")
        start = time.time()
        db = MINE(self.mine, self.mongo_uri)

        # Insert Reactions
        print("--------------- Reactions --------------")
        rxn_start = time.time()
        write_reactions_to_mine(self.reactions.values(), db)
        print(f"Wrote Reactions in {time.time() - rxn_start} seconds.")
        print("----------------------------------------\n")

        # Insert Reactions
        print("--------------- Compounds --------------")
        cpd_start = time.time()
        write_compounds_to_mine(self.compounds.values(), db, processes=processes)
        print(f"Wrote Compounds in {time.time() - cpd_start} seconds.")
        if write_core:
            cpd_start = time.time()
            print("\nWriting Core Compounds")
            write_core_compounds(
                self.compounds.values(), db, self.mine, processes=processes
            )
            print(f"Wrote Core Compounds in {time.time() - cpd_start} seconds.")
        print("----------------------------------------\n")

        if self.targets:
            print("--------------- Targets ----------------")
            target_start = time.time()
            write_targets_to_mine(self.targets.values(), db)
            print(f"Wrote Target Compounds in {time.time() - target_start}" " seconds.")
        else:
            print("No targets to write to MINE.")
        # Save operators
        # Operator values have to be calculated
        if self.operators:
            print("-------------- Operators ---------------")
            operator_start = time.time()
            # update operator rxn count
            for rxn_dict in self.reactions.values():
                for op in rxn_dict["Operators"]:
                    self.operators[op][1]["Reactions_predicted"] += 1

            db.operators.insert_many([op[1] for op in self.operators.values()])
            db.meta_data.insert_one(
                {"Timestamp": datetime.datetime.now(), "Action": "Operators Inserted"}
            )
            print(
                "Done with Operators Overall--took "
                f"{time.time() - operator_start} seconds."
            )
        print("----------------------------------------\n")

        if indexing:
            print("-------------- Indices ---------------")
            index_start = time.time()
            db.build_indexes()
            print(f"Built Indices--took {time.time() - index_start} seconds.")
            print("----------------------------------------\n")

        print("-------------- Overall ---------------")
        print(f"Finished uploading everything in {time.time() - start} sec")
        print("----------------------------------------\n")

    def _transform_helper(self, compound_smiles: List[str], processes: int) -> None:
        """Help to transform reactions external to pickaxe class

        Parameters
        ----------
        compound_smiles : List[str]
            A list of SMILES to expand.
        processes : int
            Number of processes to run.
        """

        def update_cpds_rxns(new_cpds, new_rxns):
            # Save results to self.compounds / self.reactions
            # ensuring there are no collisions and updating information
            #  if there are
            for cpd_id, cpd_dict in new_cpds.items():
                if cpd_id not in self.compounds:
                    self.compounds[cpd_id] = cpd_dict

            for rxn_id, rxn_dict in new_rxns.items():
                if rxn_id not in self.reactions:
                    self.reactions[rxn_id] = rxn_dict
                else:
                    rxn_ops = self.reactions[rxn_id]["Operators"]
                    rxn_ops = rxn_ops.union(rxn_dict["Operators"])
                    # if "Partial Operators" in self.reactions[rxn_id]:
                    #     par_ops = self.reactions[rxn_id]["Partial Operators"]
                    #     par_ops = par_ops.union(rxn_dict["Partial Operators"])

                # Update compound tracking
                for product_id in [
                    cpd_id
                    for _, cpd_id in rxn_dict["Products"]
                    if cpd_id.startswith("C")
                ]:
                    if rxn_id not in self.compounds[product_id]["Product_of"]:
                        self.compounds[product_id]["Product_of"].append(rxn_id)

                for reactant_id in [
                    cpd_id
                    for _, cpd_id in rxn_dict["Reactants"]
                    if cpd_id.startswith("C")
                ]:
                    if rxn_id not in self.compounds[reactant_id]["Reactant_in"]:
                        self.compounds[reactant_id]["Reactant_in"].append(rxn_id)

        # to pass coreactants externally
        coreactant_dict = {
            co_key: self.compounds[co_key] for _, co_key in self.coreactants.values()
        }

        new_cpds, new_rxns = transform_all_compounds_with_full(
            compound_smiles,
            self.coreactants,
            coreactant_dict,
            self.operators,
            self.generation + 1,
            self.explicit_h,
            self.kekulize,
            processes,
        )

        update_cpds_rxns(new_cpds, new_rxns)

        # if self.partial_operators:
        #     print("\nGenerating partial operators...")
        #     partial_operators = self._filter_partial_operators()
        #     if partial_operators:
        #         print("Found partial operators, applying.")
        #         # transform partial
        #         new_cpds, new_rxns = _transform_all_compounds_with_partial(
        #                                     compound_smiles, self.coreactants,
        #                                     coreactant_dict, self.operators,
        #                                     self.generation, self.explicit_h,
        #                                     processes, partial_operators
        #                             )

        #         update_cpds_rxns(new_cpds, new_rxns)
        #     else:
        #         print("No partial operators could be generated.")

    def pickle_pickaxe(self, fname: str) -> None:
        """Pickle key pickaxe items.

        Pickle pickaxe object to be loaded in later.

        Parameters
        ----------
        fname : str
            filename to save (must be .pk).
        """
        dict_to_pickle = {
            "compounds": self.compounds,
            "reactions": self.reactions,
            "operators": self.operators,
            "targets": self.targets,
        }

        with open(fname, "wb") as f:
            pickle.dump(dict_to_pickle, f)

    def load_pickled_pickaxe(self, fname: str) -> None:
        """Load pickaxe from pickle.

        Load pickled pickaxe object.

        Parameters
        ----------
        fname : str
            filename to read (must be .pk).
        """
        start_load = time.time()
        print(f"Loading {fname} pickled data.")
        with open(fname, "rb") as f:
            pickle_d = pickle.load(f)
            self.compounds = pickle_d["compounds"]
            self.reactions = pickle_d["reactions"]
            self.operators = pickle_d["operators"]
            self.targets = pickle_d["targets"]
            self.target_smiles = [target["SMILES"] for target in self.targets.values()]

            for key in pickle_d:
                var = getattr(self, key)
                if var:
                    print(f"Loaded {len(var)} {key}")

            print(f"Took {time.time() - start_load}")


if __name__ == "__main__":
    print(os.getcwd())
    # Get initial time to calculate execution time at end
    t1 = time.time()  # pylint: disable=invalid-name
    # Parse all command line arguments
    parser = ArgumentParser()  # pylint: disable=invalid-name
    # Core args
    parser.add_argument(
        "-C",
        "--coreactant_list",
        default="./tests/data/test_coreactants.tsv",
        help="Specify a list of coreactants as a .tsv",
    )
    parser.add_argument(
        "-r",
        "--rule_list",
        default="./tests/data/test_reaction_rules.tsv",
        help="Specify a list of reaction rules as a .tsv",
    )
    parser.add_argument(
        "-c",
        "--compound_file",
        default="./tests/data/test_compounds.tsv",
        help="Specify a list of starting compounds as .tsv or .csv",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="Display RDKit errors & warnings",
    )
    # parser.add_argument('--bnice', action='store_true', default=False,
    #                     help="Set several options to enable compatibility "
    #                          "with bnice operators.")
    parser.add_argument(
        "-H",
        "--explicit_h",
        action="store_true",
        default=False,
        help="Specify explicit hydrogen for use in reaction rules.",
    )
    parser.add_argument(
        "-k",
        "--kekulize",
        action="store_true",
        default=False,
        help="Specify whether to kekulize compounds.",
    )
    parser.add_argument(
        "-n",
        "--neutralise",
        action="store_true",
        default=True,
        help="Specify whether to neturalise compounds.",
    )

    parser.add_argument(
        "-m",
        "--processes",
        default=1,
        type=int,
        help="Set the max number of processes.",
    )

    parser.add_argument(
        "-g",
        "--generations",
        default=1,
        type=int,
        help="Set the numbers of time to apply the reaction rules to the compound set.",
    )

    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        default=True,
        help="Silence warnings about imbalanced reactions",
    )

    parser.add_argument(
        "-s", "--smiles", default=None, help="Specify a starting compound SMILES."
    )

    # Result args
    parser.add_argument(
        "-p",
        "--pruning_whitelist",
        default=None,
        help="Specify a list of target compounds to prune reaction network down to.",
    )
    parser.add_argument(
        "-o",
        "--output_dir",
        default=None,
        help="The directory in which to write files.",
    )
    parser.add_argument(
        "-d",
        "--database",
        default=None,
        help="The URI of the database in which to store "
        "output. If not specified, data is written "
        "as tsv files",
    )
    parser.add_argument(
        "-u",
        "--mongo_uri",
        default="mongodb://localhost:27017",
        help="The URI of the mongo database to connect to. Defaults to"
        " mongodb://localhost:27017",
    )
    parser.add_argument(
        "-i",
        "--image_dir",
        default=None,
        help="Specify a directory to store images of all created compounds",
    )
    parser.add_argument(
        "-wc",
        "--write_core",
        default=False,
        help="Whether or not to write results into core database.",
    )

    OPTIONS = parser.parse_args()

    if not any([OPTIONS.database, OPTIONS.output_dir]):
        exit("No output selected, terminating run.")

    pk = Pickaxe(
        coreactant_list=OPTIONS.coreactant_list,
        rule_list=OPTIONS.rule_list,
        errors=OPTIONS.verbose,
        explicit_h=OPTIONS.explicit_h,
        kekulize=OPTIONS.kekulize,
        neutralise=OPTIONS.neutralise,
        image_dir=OPTIONS.image_dir,
        quiet=OPTIONS.quiet,
        database=OPTIONS.database,
        mongo_uri=OPTIONS.mongo_uri,
    )
    # Create a directory for image output file if it doesn't already exist
    if OPTIONS.image_dir and not os.path.exists(OPTIONS.image_dir):
        os.mkdir(OPTIONS.image_dir)
    # If starting compound specified as SMILES string, then add it
    if OPTIONS.smiles:
        pk._add_compound("Start", OPTIONS.smiles, cpd_type="Starting Compound")
    else:
        pk.load_compound_set(compound_file=OPTIONS.compound_file)
    # Generate reaction network
    pk.transform_all(processes=OPTIONS.processes, generations=OPTIONS.generations)
    if OPTIONS.pruning_whitelist:
        mols = [
            pk._mol_from_dict(line)
            for line in utils.file_to_dict_list(OPTIONS.pruning_whitelist)
        ]
        pk.prune_network([utils.get_compound_hash(x) for x in mols if x])

    if OPTIONS.output_dir:
        pk.assign_ids()
        pk.write_compound_output_file(OPTIONS.output_dir + "/compounds.tsv")
        pk.write_reaction_output_file(OPTIONS.output_dir + "/reactions.tsv")

    if OPTIONS.database:
        pk.save_to_mine(processes=OPTIONS.processes, write_core=OPTIONS.write_core)

    print(f"Execution took {time.time() - t1} seconds.")



"""Tests for rules.py using pytest."""


import json
from pathlib import Path

from minedatabase import pickaxe
from minedatabase.rules import BNICE, metacyc_generalized, metacyc_intermediate


file_path = Path(__file__)
file_dir = file_path.parent

with open((file_dir / "../data/test_rules/rules_to_assert.json"), "r") as f:
    rule_assert_dict = json.load(f)


def test_metacyc_generalized_full():
    rule_list, correactant_list, rule_name = metacyc_generalized()
    pk = pickaxe.Pickaxe(rule_list=rule_list, coreactant_list=correactant_list)

    assert rule_name == "Metacyc_generalized"
    assert len(pk.operators) == 1216
    assert len(pk.coreactants) == 45

    assert (
        pk.operators["rule0001"][1]["SMARTS"]
        == rule_assert_dict["Metacyc_generalized_rule0001"]
    )


def test_metacyc_generalized_specify_number():
    rule_list, correactant_list, rule_name = metacyc_generalized(n_rules=10)
    pk = pickaxe.Pickaxe(rule_list=rule_list, coreactant_list=correactant_list)

    assert rule_name == "Metacyc_generalized_10_rules"
    assert len(pk.operators) == 10
    assert len(pk.coreactants) == 45

    assert (
        pk.operators["rule0001"][1]["SMARTS"]
        == rule_assert_dict["Metacyc_generalized_rule0001"]
    )


def test_metacyc_generalized_specify_fraction():
    rule_list, correactant_list, rule_name = metacyc_generalized(fraction_coverage=0.9)
    pk = pickaxe.Pickaxe(rule_list=rule_list, coreactant_list=correactant_list)

    assert rule_name == "Metacyc_generalized_0,9_fraction_coverage"
    assert len(pk.operators) == 413
    assert len(pk.coreactants) == 45

    assert (
        pk.operators["rule0001"][1]["SMARTS"]
        == rule_assert_dict["Metacyc_generalized_rule0001"]
    )


def test_metacyc_exclude():
    rule_list, correactant_list, rule_name = metacyc_generalized(
        fraction_coverage=0.9, exclude_containing=["aromatic", "halogen"]
    )
    pk = pickaxe.Pickaxe(rule_list=rule_list, coreactant_list=correactant_list)

    assert rule_name == "Metacyc_generalized_0,9_fraction_coverage_with_exclusion"
    assert len(pk.operators) == 839
    assert len(pk.coreactants) == 45


def test_metacyc_include():
    rule_list, correactant_list, rule_name = metacyc_generalized(
        fraction_coverage=0.9, include_containing=["aromatic", "halogen"]
    )
    pk = pickaxe.Pickaxe(rule_list=rule_list, coreactant_list=correactant_list)

    assert rule_name == "Metacyc_generalized_0,9_fraction_coverage_with_inclusion"
    assert len(pk.operators) == 377
    assert len(pk.coreactants) == 45


def test_metacyc_intermediate():
    rule_list, correactant_list, rule_name = metacyc_intermediate()
    pk = pickaxe.Pickaxe(rule_list=rule_list, coreactant_list=correactant_list)

    assert rule_name == "Metacyc_intermediate"
    assert len(pk.operators) == 7154
    assert len(pk.coreactants) == 45

    assert (
        pk.operators["rule0001_0167"][1]["SMARTS"]
        == rule_assert_dict["Metacyc_intermediate_rule0001_0167"]
    )


def test_metacyc_intermediate_specify_number():
    rule_list, correactant_list, rule_name = metacyc_intermediate(n_rules=20)
    pk = pickaxe.Pickaxe(rule_list=rule_list, coreactant_list=correactant_list)

    assert rule_name == "Metacyc_intermediate_20_rules"
    assert len(pk.operators) == 20
    assert len(pk.coreactants) == 45


def test_metacyc_intermediate_specify_fraction():
    rule_list, correactant_list, rule_name = metacyc_intermediate(fraction_coverage=0.2)
    pk = pickaxe.Pickaxe(rule_list=rule_list, coreactant_list=correactant_list)

    assert rule_name == "Metacyc_intermediate_0,2_fraction_coverage"
    assert len(pk.operators) == 75
    assert len(pk.coreactants) == 45


def test_metacyc_intermediate_exclude():
    rule_list, correactant_list, rule_name = metacyc_intermediate(
        fraction_coverage=0.9, exclude_containing=["aromatic", "halogen"]
    )
    pk = pickaxe.Pickaxe(rule_list=rule_list, coreactant_list=correactant_list)

    assert rule_name == "Metacyc_intermediate_0,9_fraction_coverage_with_exclusion"
    assert len(pk.operators) == 5775
    assert len(pk.coreactants) == 45


def test_metacyc_intermediate_include():
    rule_list, correactant_list, rule_name = metacyc_intermediate(
        fraction_coverage=0.9, include_containing=["halogen"]
    )
    pk = pickaxe.Pickaxe(rule_list=rule_list, coreactant_list=correactant_list)

    assert rule_name == "Metacyc_intermediate_0,9_fraction_coverage_with_inclusion"
    assert len(pk.operators) == 67
    assert len(pk.coreactants) == 45


def test_BNICE():
    rule_list, correactant_list, rule_name = BNICE()
    pk = pickaxe.Pickaxe(rule_list=rule_list, coreactant_list=correactant_list)

    assert rule_name == "BNICE"
    assert len(pk.operators) == 250
    assert len(pk.coreactants) == 33

    assert pk.operators["1.1.1_01"][1]["SMARTS"] == rule_assert_dict["BNICE_1.1.1_01"]



# from minedatabase.filters.feasibility import ReactionFeasibilityFilter
from minedatabase.filters.metabolomics import MetabolomicsFilter
from minedatabase.filters.property import AtomicCompositionFilter, MWFilter
from minedatabase.filters.similarity import (
    MCSFilter,
    SimilarityFilter,
    SimilaritySamplingFilter,
)

# from minedatabase.filters.thermodynamics import ThermoFilter
# from minedatabase.filters.toxicity import ToxicityFilter



"""Tests for pickaxe.py using pytest."""
# pylint: disable=redefined-outer-name
# pylint: disable=protected-access

import hashlib
import os
import re
import subprocess
from filecmp import cmp
from pathlib import Path

import pymongo
import pytest
from pymongo.errors import ServerSelectionTimeoutError
from rdkit.Chem import AllChem

from minedatabase import pickaxe
from minedatabase.databases import MINE


try:
    client = pymongo.MongoClient(ServerSelectionTimeoutMS=20)
    client.server_info()
    del client
    is_mongo = True
except ServerSelectionTimeoutError as err:
    is_mongo = False
valid_db = pytest.mark.skipif(not is_mongo, reason="No MongoDB Connection")

file_path = Path(__file__)
file_dir = file_path.parent

DATA_DIR = (file_dir / "../data/").resolve()


def purge(directory, pattern):
    """Delete all files in a directory matching a regex pattern."""
    for filename in os.listdir(directory):
        if re.search(pattern, filename):
            os.remove(os.path.join(directory, filename))


def delete_database(name):
    """Delete database."""
    mine = MINE(name)
    mine.client.drop_database(name)
    mine.client.close()


def test_cofactor_loading(pk):
    """Test loading cofactors.

    GIVEN a default Pickaxe object
    WHEN cofactors are loaded into the Pickaxe object in its creation
    THEN make sure those cofactors were loaded correctly
    """
    c_id = "X73bc8ef21db580aefe4dbc0af17d4013961d9d17"

    assert c_id in pk.compounds
    assert pk.compounds[c_id]["Formula"] == "H2O"
    assert pk.compounds[c_id]["Type"] == "Coreactant"
    assert isinstance(pk.coreactants["Water"][0], AllChem.Mol)
    assert pk.coreactants["Water"][1][0] == "X"


def test_reaction_rule_loading(default_rule):
    """Test loading rules.

    GIVEN a reaction rule dict
    WHEN reaction rules are loaded during Pickaxe object initialization
    THEN make sure it is formatted correctly
    """
    assert isinstance(default_rule[0], AllChem.ChemicalReaction)
    assert isinstance(default_rule[1], dict)
    assert default_rule[1]["Reactants"] == ["ATP", "Any"]
    assert "Products" in default_rule[1]
    assert "Comments" in default_rule[1]


def test_compound_loading(pk):
    """Test loading compounds.

    GIVEN a default Pickaxe object
    WHEN compounds are loaded
    THEN check that they are loaded correctly
    """
    compound_smiles = pk.load_compound_set(
        compound_file=file_dir / "../data/test_compounds.tsv"
    )
    assert len(compound_smiles) == 14


def test_transform_all(default_rule, smiles_dict, coreactant_dict):
    """Test transform function.

    GIVEN a set of rules and starting compounds
    WHEN we run pickaxe to predict potential transformations
    THEN make sure all expected transformations are predicted
    """
    pk = pickaxe.Pickaxe(errors=False, explicit_h=True)
    pk._load_coreactant(coreactant_dict["ATP"])
    pk._load_coreactant(coreactant_dict["ADP"])
    pk._add_compound(
        smiles_dict["FADH"], smiles_dict["FADH"], cpd_type="Starting Compound"
    )
    pk.operators["2.7.1.a"] = default_rule
    pk.transform_all(generations=2)
    assert len(pk.compounds) == 31
    assert len(pk.reactions) == 49
    comp_gens = set([x["Generation"] for x in pk.compounds.values()])
    assert comp_gens == {0, 1, 2}


def test_compound_output_writing(pk_transformed):
    """Test compound output writing.

    GIVEN a Pickaxe object with predicted transformations
    WHEN all compounds (including predicted) are written to an output file
    THEN make sure they are correctly written, and that they are all present
    """

    with open(file_dir / "../data/testcompoundsout.tsv", "rb") as infile:
        expected = hashlib.sha256(infile.read()).hexdigest()
    pk_transformed.write_compound_output_file(file_dir / "../data/testcompoundsout.tsv")
    assert os.path.exists(file_dir / "../data/testcompoundsout_new.tsv")
    try:
        with open(file_dir / "../data/testcompoundsout_new.tsv", "rb") as infile:
            output_compounds = hashlib.sha256(infile.read()).hexdigest()
        assert expected == output_compounds
    finally:
        os.remove(file_dir / "../data/testcompoundsout_new.tsv")


def test_reaction_output_writing(pk_transformed):
    """Test writing reaction output.

    GIVEN a Pickaxe object with predicted transformations
    WHEN all reactions (including predicted) are written to an output file
    THEN make sure they are correctly written, and that they are all present
    """
    with open(file_dir / "../data/testreactionsout.tsv", "rb") as infile:
        expected = hashlib.sha256(infile.read()).hexdigest()
    pk_transformed.write_reaction_output_file(file_dir / "../data/testreactionsout.tsv")
    assert os.path.exists(file_dir / "../data/testreactionsout_new.tsv")
    try:
        with open(file_dir / "../data/testreactionsout_new.tsv", "rb") as infile:
            output_compounds = hashlib.sha256(infile.read()).hexdigest()
        assert expected == output_compounds
    finally:
        os.remove(file_dir / "../data/testreactionsout_new.tsv")


def test_multiprocessing(pk, smiles_dict, coreactant_dict):
    """Test multiprocessing.

    GIVEN a Pickaxe object
    WHEN we use multiprocessing to enumerate predicted reactions
    THEN make sure those predictions are correct
    """
    pk._load_coreactant(coreactant_dict["ATP"])
    pk._load_coreactant(coreactant_dict["ADP"])
    pk._add_compound("FADH", smiles_dict["FADH"], cpd_type="Starting Compound")
    pk.transform_all(generations=2, processes=2)

    assert len(pk.compounds) == 67
    assert len(pk.reactions) == 49
    comp_gens = set([x["Generation"] for x in pk.compounds.values()])
    assert comp_gens == {0, 1, 2}


def test_pruning(default_rule, smiles_dict, coreactant_dict):
    """Test pruning network to targets.

    GIVEN a Pickaxe expansion
    WHEN that expansion is pruned via Pickaxe.prune_network()
    THEN make sure that the pruned compounds no longer exist in the network
    """

    pk = pickaxe.Pickaxe(explicit_h=True)
    pk.operators["2.7.1.a"] = default_rule
    pk._load_coreactant(coreactant_dict["ATP"])
    pk._load_coreactant(coreactant_dict["ADP"])
    pk._add_compound("FADH", smiles_dict["FADH"], cpd_type="Starting Compound")
    pk.transform_all(generations=2)

    ids = [
        "C89d19c432cbe8729c117cfe50ff6ae4704a4e6c1",
        "C750e93db23dd3f796ffdf9bdefabe32b10710053",
        "C41",
    ]
    pk.prune_network(ids)
    pk.assign_ids()

    DATA_DIR = (file_dir / "../data").resolve()
    pk.write_compound_output_file(DATA_DIR / "pruned_comps.tsv")
    pk.write_reaction_output_file(DATA_DIR / "pruned_rxns.tsv")
    assert os.path.exists(DATA_DIR / "pruned_comps_new.tsv")
    assert os.path.exists(DATA_DIR / "pruned_rxns_new.tsv")
    try:
        assert cmp(DATA_DIR / "pruned_comps.tsv", DATA_DIR / "pruned_comps_new.tsv")
        assert cmp(DATA_DIR / "pruned_rxns.tsv", DATA_DIR / "pruned_rxns_new.tsv")
    finally:
        os.remove((DATA_DIR / "pruned_comps_new.tsv").resolve())
        os.remove((DATA_DIR / "pruned_rxns_new.tsv").resolve())


def test_target_generation(default_rule, smiles_dict, coreactant_dict):
    """Test generating a target from starting compounds."""
    pk = pickaxe.Pickaxe(explicit_h=True)
    pk.operators["2.7.1.a"] = default_rule
    pk._load_coreactant(coreactant_dict["ATP"])
    pk._load_coreactant(coreactant_dict["ADP"])
    pk._add_compound("FADH", smiles_dict["FADH"], cpd_type="Starting Compound")
    pk.load_targets(file_dir / "../data/test_targets.csv")
    pk.transform_all(generations=2)
    pk.prune_network_to_targets()

    assert "C11088915f64b93293e70af9c3b7822a4f131225d" in pk.compounds
    assert len(pk.reactions) == 4
    assert len(pk.compounds) == 6


@valid_db
def test_save_as_mine(default_rule, smiles_dict, coreactant_dict):
    """Test saving compounds to database.

    GIVEN a Pickaxe expansion
    WHEN that expansion is saved as a MINE DB in the MongoDB
    THEN make sure that all features are saved in the MongoDB as expected
    """
    DATA_DIR = (file_dir / "../data").resolve()
    delete_database("MINE_test")
    pk = pickaxe.Pickaxe(database="MINE_test", image_dir=DATA_DIR, explicit_h=True)
    pk.operators["2.7.1.a"] = default_rule
    pk._load_coreactant(coreactant_dict["ATP"])
    pk._load_coreactant(coreactant_dict["ADP"])
    pk._add_compound("FADH", smiles_dict["FADH"], cpd_type="Starting Compound")
    pk.transform_all(generations=2)
    pk.save_to_mine(processes=1)
    mine_db = MINE("MINE_test")

    try:

        assert mine_db.compounds.estimated_document_count() == 31
        assert mine_db.reactions.estimated_document_count() == 49
        assert mine_db.operators.estimated_document_count() == 1
        assert mine_db.operators.find_one()["Reactions_predicted"] == 49
        assert os.path.exists(
            DATA_DIR / "X9c29f84930a190d9086a46c344020283c85fb917.svg"
        )
        start_comp = mine_db.compounds.find_one({"Type": "Starting Compound"})
        assert len(start_comp["Reactant_in"]) > 0
        # Don't track sources of coreactants
        coreactant = mine_db.compounds.find_one({"Type": "Coreactant"})
        assert "Product_of" not in coreactant
        assert "Reactant_in" not in coreactant
        product = mine_db.compounds.find_one({"Generation": 2})
        assert len(product["Product_of"]) > 0
        assert product["Type"] == "Predicted"
    finally:
        delete_database("MINE_test")
        purge(DATA_DIR, r".*\.svg$")


@valid_db
def test_save_target_mine(default_rule, smiles_dict, coreactant_dict):
    """Test saving the target run to a MINE."""
    delete_database("MINE_test")
    pk = pickaxe.Pickaxe(database="MINE_test", explicit_h=True)
    pk.operators["2.7.1.a"] = default_rule
    pk._load_coreactant(coreactant_dict["ATP"])
    pk._load_coreactant(coreactant_dict["ADP"])
    pk._add_compound("FADH", smiles_dict["FADH"], cpd_type="Starting Compound")
    pk.load_targets(file_dir / "../data/test_targets.csv")
    pk.transform_all(generations=2)
    pk.prune_network_to_targets()

    pk.save_to_mine()
    mine_db = MINE("MINE_test")

    try:
        assert mine_db.compounds.estimated_document_count() == 6
        assert mine_db.reactions.estimated_document_count() == 4
        assert mine_db.operators.estimated_document_count() == 1
        assert mine_db.operators.find_one()["Reactions_predicted"] == 4
        start_comp = mine_db.target_compounds.find_one()
        assert start_comp["InChI_key"] == "RYNUDNWPSBJQQY-UHFFFAOYSA-N"
        assert all([i in start_comp.keys() for i in ["_id", "SMILES", "InChI_key"]])
    finally:
        delete_database("MINE_test")


@valid_db
def test_database_already_exists(default_rule, smiles_dict, coreactant_dict):
    """Test database collision.

    GIVEN an existing MINE
    WHEN a new pickaxe object is defined
    THEN make sure program exits with database collision
    """
    delete_database("MINE_test")
    pk = pickaxe.Pickaxe(database="MINE_test")
    pk.operators["2.7.1.a"] = default_rule
    pk._load_coreactant(coreactant_dict["ATP"])
    pk._load_coreactant(coreactant_dict["ADP"])
    pk._add_compound("FADH", smiles_dict["FADH"], cpd_type="Starting Compound")
    pk.transform_all(generations=2)
    pk.save_to_mine(processes=1)

    try:
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            pk = pickaxe.Pickaxe(database="MINE_test")
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == (
            "Exiting due to database name collision."
        )
    finally:
        delete_database("MINE_test")


def test_pickle(coreactant_dict, smiles_dict, default_rule):
    """Test pickling of pickaxe objects."""
    pickle_path = Path("test_pickle.pk")

    pk = pickaxe.Pickaxe(errors=False, explicit_h=True)
    pk._load_coreactant(coreactant_dict["ATP"])
    pk._load_coreactant(coreactant_dict["ADP"])
    pk._add_compound(
        smiles_dict["FADH"], smiles_dict["FADH"], cpd_type="Starting Compound"
    )
    pk.operators["2.7.1.a"] = default_rule

    pk.transform_all(generations=2)

    pk.pickle_pickaxe(pickle_path)
    del pk
    pk = pickaxe.Pickaxe(errors=False)
    pk.load_pickled_pickaxe(pickle_path)

    assert len(pk.compounds) == 31
    assert len(pk.reactions) == 49
    comp_gens = set([x["Generation"] for x in pk.compounds.values()])
    assert comp_gens == {0, 1, 2}

    pickle_path.unlink()


def test_local_cli():
    """Test command line interface writing locally.

    GIVEN the pickaxe CLI
    WHEN pickaxe is run from the command line
    THEN make sure it exits with exit code 0 (no errors)
    """
    os.chdir(file_dir / "../data/../..")
    rc = subprocess.call(
        f"python minedatabase/pickaxe.py -o tests/ -r tests/data/test_cd_rxn_rule.tsv",
        shell=True,
    )
    assert not rc
    purge(DATA_DIR / "..", r".*\.tsv$")


@valid_db
def test_mongo_cli():
    """Test command line interface writing to mongo."""
    mine = MINE("tests")
    os.chdir(file_dir / "../data/../..")
    rc = subprocess.call(
        "python minedatabase/pickaxe.py -d tests -r tests/data/test_cd_rxn_rule.tsv",
        shell=True,
    )
    assert not rc

    try:
        assert mine.compounds.estimated_document_count() == 51
    finally:
        mine.client.drop_database("tests")

    purge(file_dir / "..", r".*\.svg$")



"""Tests for pickaxe.py using pytest."""
from pathlib import Path

import pytest

from minedatabase.filters import (
    AtomicCompositionFilter,
    MCSFilter,
    MetabolomicsFilter,
    MWFilter,
    SimilarityFilter,
    SimilaritySamplingFilter,
)

file_path = Path(__file__)
file_dir = file_path.parent
DATA_DIR = (file_dir / "../data/").resolve()

# check for eQ
sqlite_loc = file_dir / "../compounds.sqlite"

# Thermodynamics
loaded_db = False

if sqlite_loc.is_file():
    try:
        from equilibrator_api import Q_

        from minedatabase.filters.thermodynamics import ThermoFilter
        from minedatabase.thermodynamics import Thermodynamics

        thermo = Thermodynamics()
        try:
            thermo.load_thermo_from_postgres()
            loaded_db = True
        except:
            try:
                thermo.load_thermo_from_sqlite()
                loaded_db = True
            except:
                pass
    except:
        pass


def test_similarity_cutoff_single(pk_target):
    """Test similarity cutoff filter"""
    tani_threshold = 0.5
    _filter = SimilarityFilter(
        crit_similarity=tani_threshold, increasing_similarity=False
    )
    pk_target.filters.append(_filter)
    pk_target.transform_all(generations=2)

    assert len(pk_target.compounds) == 436
    assert (
        pk_target.compounds["C779bfa0d747509f0499664b390657a336edec104"]["Expand"]
        == True
    )


@pytest.mark.skip("Heisenbug Test")
def test_filter_after(pk_target):
    """Test similarity cutoff filter"""
    tani_threshold = 0.5
    _filter = SimilarityFilter(
        crit_similarity=tani_threshold, increasing_similarity=False
    )
    pk_target.filter_after_final_gen = True
    pk_target.filters.append(_filter)
    pk_target.transform_all(generations=2)

    assert len(pk_target.compounds) == 257
    assert (
        pk_target.compounds["C779bfa0d747509f0499664b390657a336edec104"]["Expand"]
        == False
    )


def test_similarity_cutoff_multi(pk_target):
    """Test similarity cutoff filter"""
    tani_threshold = [0, 0.3, 0.5]
    _filter = SimilarityFilter(
        crit_similarity=tani_threshold, increasing_similarity=False
    )
    pk_target.filters.append(_filter)
    pk_target.transform_all(generations=2)

    assert len(pk_target.compounds) == 1358
    assert (
        pk_target.compounds["C779bfa0d747509f0499664b390657a336edec104"]["Expand"]
        == True
    )


def test_similarity_cutoff_multi_short_list(pk_target):
    """Test similarity filter when the tani_threshold is shorter than generations."""
    tani_threshold = [0.5]
    _filter = SimilarityFilter(
        crit_similarity=tani_threshold, increasing_similarity=False
    )
    pk_target.filters.append(_filter)
    pk_target.transform_all(generations=2)

    assert len(pk_target.compounds) == 436
    assert (
        pk_target.compounds["C779bfa0d747509f0499664b390657a336edec104"]["Expand"]
        == True
    )


def test_similarity_no_targets(pk_target):
    pk_target.target_smiles = []
    tani_threshold = 0.5
    _filter = SimilarityFilter(
        crit_similarity=tani_threshold, increasing_similarity=False
    )

    pk_target.filters.append(_filter)
    pk_target.transform_all(generations=2)

    assert len(pk_target.compounds) == 1740
    assert (
        pk_target.compounds["C779bfa0d747509f0499664b390657a336edec104"]["Expand"]
        == True
    )


def test_similarity_sample_default_weight(pk_target):
    """Test similarity cutoff filter"""
    _filter = SimilaritySamplingFilter(sample_size=10, weight=None)
    pk_target.filters.append(_filter)
    pk_target.transform_all(generations=2)

    # Filter must return less compounds than non-filter
    # Non-deterministic results, so no exact value can be used
    assert len(pk_target.compounds) < 1452


def test_similarity_sample_user_weight(pk_target):
    """Test similarity cutoff filter"""

    def weight(T):
        return T ** 4

    _filter = SimilaritySamplingFilter(sample_size=10, weight=weight)
    pk_target.filters.append(_filter)
    pk_target.transform_all(generations=2)

    # Filter must return less compounds than non-filter
    # Non-deterministic results, so no exact value can be used
    assert len(pk_target.compounds) < 1452


def test_similarity_sample_morgan(pk_target):
    """Test overwriting defaults"""

    fingerprint_method = "Morgan"
    fingerprint_args = {"radius": 2}

    _filter = SimilaritySamplingFilter(
        sample_size=10,
        fingerprint_method=fingerprint_method,
        fingerprint_args=fingerprint_args,
    )
    pk_target.filters.append(_filter)
    pk_target.transform_all(generations=2)

    # Filter must return less compounds than non-filter
    # Non-deterministic results, so no exact value can be used
    assert len(pk_target.compounds) < 1452 and len(pk_target.compounds) > 100


def test_similarity_sample_dice(pk_target):
    """Test overwriting defaults"""

    fingerprint_method = "Morgan"
    fingerprint_args = {"radius": 2}
    similarity_method = "Dice"

    _filter = SimilaritySamplingFilter(
        sample_size=10,
        fingerprint_method=fingerprint_method,
        fingerprint_args=fingerprint_args,
        similarity_method=similarity_method,
    )

    pk_target.filters.append(_filter)
    pk_target.transform_all(processes=2, generations=2)

    # Filter must return less compounds than non-filter
    # Non-deterministic results, so no exact value can be used
    assert len(pk_target.compounds) < 1452 and len(pk_target.compounds) > 100


def test_similarity_sample_multiprocess(pk_target):
    """Test similarity cutoff filter"""

    def weight(T):
        return T ** 4

    _filter = SimilaritySamplingFilter(sample_size=10, weight=weight)
    pk_target.react_targets = True
    pk_target.filters.append(_filter)
    pk_target.transform_all(processes=2, generations=2)

    # Filter must return less compounds than non-filter
    # Non-deterministic results, so no exact value can be used
    assert len(pk_target.compounds) < 1452


def test_MCS_list(pk_target):
    """Test similarity cutoff filter"""
    MCS_threshold = [0.1, 0.5]
    _filter = MCSFilter(crit_mcs=MCS_threshold)
    pk_target.filters.append(_filter)
    pk_target.transform_all(generations=2)

    assert len(pk_target.compounds) == 382


@pytest.mark.skipif(not loaded_db, reason="No eQuilibrator DB found.")
def test_thermo_phys(pk_target):
    """Test thermo cutoff for physiological"""
    _filter = ThermoFilter(physiological=True)
    pk_target.filters.append(_filter)
    pk_target.transform_all(generations=1)

    assert True


def test_met_filter_mass(pk_target):
    """Test MetabolomicsFilter output without RT predictor."""
    metabolomics_data_path = DATA_DIR / "test_metabolomics/test_metabolomics_data.csv"
    met_filter = MetabolomicsFilter(
        filter_name="test_metabolomics_filter",
        met_data_name="test_metabolomics_data",
        met_data_path=metabolomics_data_path,
        possible_adducts=["[M+H]+", "[M-H]-"],
        mass_tolerance=0.001,
    )
    pk_target.filters.append(met_filter)
    pk_target.transform_all(generations=2)

    gen1_cpds = [
        pk_target.compounds[cpd]
        for cpd in pk_target.compounds
        if pk_target.compounds[cpd]["Generation"] == 1
    ]

    assert len(gen1_cpds) == 1
    assert gen1_cpds[0]["Matched_Peak_IDs"] == ["Test3"]



"""Tests for utils.py using pytest."""

import os
from collections import OrderedDict
from pathlib import Path

from rdkit.DataStructs.cDataStructs import ExplicitBitVect

import minedatabase.utils as utils


file_path = Path(__file__)
file_dir = file_path.parent

DATA_DIR = (file_dir / "../data/").resolve()


def test_get_compound_hash():
    """Test compound to hash."""
    assert utils.get_compound_hash("CCO", "Coreactant") == (
        "Xa41fe8492d86f214ba494e3d04da2f0854c0e2ea",
        "LFQSCWFLJHTTHZ",
    )
    assert utils.get_compound_hash("CCO", "Predicted") == (
        "Ca41fe8492d86f214ba494e3d04da2f0854c0e2ea",
        "LFQSCWFLJHTTHZ-UHFFFAOYSA-N",
    )
    assert utils.get_compound_hash("CCO", "Starting Compound") == (
        "Ca41fe8492d86f214ba494e3d04da2f0854c0e2ea",
        "LFQSCWFLJHTTHZ-UHFFFAOYSA-N",
    )


def test_get_fp():
    smiles = "C([C@@H]1[C@H]([C@@H]([C@H]([C@H](O1)O)O)O)O)O"
    assert type(utils.get_fp(smiles)) == ExplicitBitVect


def test_get_compound_hash_two_blocks():
    smiles = "C([C@@H]1[C@H]([C@@H]([C@H]([C@H](O1)O)O)O)O)O"
    assert utils.get_compound_hash(smiles, "Starting Compound", inchi_blocks=2) == (
        "Cf95a3c17f908e427c3127b4e8c3d8575c286d6ce",
        "WQZGKKKJIJFFOK-DVKNGEFBSA-N",
    )

    assert utils.get_compound_hash(smiles, "Starting Compound", inchi_blocks=1) == (
        "C9ab1a08d72c90a8167d1f3a668d8f1138e534a07",
        "WQZGKKKJIJFFOK-DVKNGEFBSA-N",
    )


def test_file_to_dict_list():
    """
    GIVEN compound files of different types (.tsv, .csv, and .json)
    WHEN the file contents are converted to a list of compound dicts
    THEN check that the list of compound dicts is produced as expected
    """
    res_7 = OrderedDict(
        [
            ("id", "cpd01211"),
            ("abbreviation", "tcynt"),
            ("name", "Thiocyanate"),
            ("formula", "CNS"),
            ("mass", "58"),
            ("source", "ModelSEED"),
            ("structure", "InChI=1S/CHNS/c2-1-3/h3H"),
            ("charge", "-1"),
            ("is_core", "1"),
            ("is_obsolete", "0"),
            ("linked_compound", "null"),
            ("is_cofactor", "0"),
            ("deltag", "22.2"),
            ("deltagerr", "5.68687"),
            ("pka", "3:0.5"),
            ("pkb", ""),
            ("abstract_compound", "null"),
            ("comprised_of", "null"),
            ("aliases", "null"),
        ]
    )

    filenames = ["test_compounds.tsv", "test_compounds.csv", "test_compounds.json"]

    for file in filenames:
        res = utils.file_to_dict_list(DATA_DIR / file)
        assert len(res) == 15
        assert res[7] == res_7



import abc
import time
from copy import copy
from typing import List, Set

import rdkit.rdBase as rkrb
import rdkit.RDLogger as rkl

from minedatabase.pickaxe import Pickaxe


logger = rkl.logger()
logger.setLevel(rkl.ERROR)
rkrb.DisableLog("rdApp.error")


class Filter(metaclass=abc.ABCMeta):
    """Abstract base class used to generate filters.

    The Filter class provides the framework for interaction with pickaxe expansions.
    Each filter subclass must inherit properties from the Filter class.
    All subclasses must implement properties and methods decorated with
    @abc.abstractmethod. Feel free to override other non-private methods as
    well, such as _pre_print() and _post_print().
    """

    @property
    @abc.abstractmethod
    def filter_name(self) -> str:
        """Obtain name of filter."""
        pass

    @abc.abstractmethod
    def _choose_items_to_filter(self, pickaxe: Pickaxe, processes: int) -> Set[str]:
        """Return list of compounds to remove from pickaxe object.

        Parameters
        ----------
        pickaxe : Pickaxe
            Instance of Pickaxe being used to expand and filter the network.
        processes : int
            The number of processes to use, by default 1.
        generation : int
            Which generation the expansion is in.
        """
        pass

    def apply_filter(
        self,
        pickaxe: Pickaxe,
        processes: int = 1,
        generation: int = 0,
        print_on: bool = True,
    ) -> None:
        """Apply filter from Pickaxe object.

        Parameters
        ----------
        pickaxe : Pickaxe
            The Pickaxe object to filter.
        processes : int
            The number of processes to use, by default 1.
        print_on : bool
            Whether or not to print filtering results.
        """
        time_sample = time.time()

        self.generation = generation

        if print_on:
            n_total = self._get_n(pickaxe, "total")
            self._pre_print_header(pickaxe)
            self._pre_print()

        compound_ids_to_check, reaction_ids_to_check = self._choose_items_to_filter(
            pickaxe, processes
        )

        self._apply_filter_results(
            pickaxe, compound_ids_to_check, reaction_ids_to_check
        )

        if print_on:
            n_filtered = self._get_n(pickaxe, "filtered")
            self._post_print(pickaxe, n_total, n_filtered, time_sample)
            self._post_print_footer(pickaxe)

    def _pre_print_header(self, pickaxe: Pickaxe) -> None:
        """Print header before filtering.

        Parameters
        ----------
        pickaxe : Pickaxe
            Instance of Pickaxe being used to expand and filter the network.
        """
        print("----------------------------------------")
        print(f"Filtering Generation {pickaxe.generation}\n")

    def _pre_print(self) -> None:
        """Print filter being applied."""
        print(f"Applying filter: {self.filter_name}")

    def _post_print(
        self, pickaxe: Pickaxe, n_total: int, n_filtered: int, time_sample: float
    ) -> None:
        """Print results of filtering.

        Parameters
        ----------
        pickaxe : Pickaxe
            Instance of Pickaxe being used to expand and filter the network.
            Unused here, but may be useful in your implementation.
        n_total : int
            Total number of compounds.
        n_filtered : int
            Number of compounds remaining after filtering.
        times_sample : float
            Time in seconds from time.time().
        """
        print(
            f"{n_filtered} of {n_total} compounds remain after applying "
            f"filter: {self.filter_name}"
            f"--took {round(time.time() - time_sample, 2)}s.\n"
        )

    def _post_print_footer(self, pickaxe: Pickaxe) -> None:
        """Print end of filtering.

        Parameters
        ----------
        pickaxe : Pickaxe
            Instance of Pickaxe being used to expand and filter the network.
        """
        print(f"Done filtering Generation {pickaxe.generation}")
        print("----------------------------------------\n")

    def _get_n(self, pickaxe: Pickaxe, n_type: str) -> int:
        """Get current number of compounds to be filtered.

        Parameters
        ----------
        pickaxe : Pickaxe
            Instance of Pickaxe being used to expand and filter the network.
        n_type : str
            Whether to return "total" number of "filtered" number of compounds.

        Returns
        -------
        n : int
            Either the total or filtered number of compounds.
        """
        n = 0
        for cpd_dict in pickaxe.compounds.values():
            is_in_current_gen = cpd_dict["Generation"] == pickaxe.generation
            is_predicted_compound = cpd_dict["_id"].startswith("C")
            if is_in_current_gen and is_predicted_compound:
                if n_type == "total":
                    n += 1
                elif n_type == "filtered" and cpd_dict["Expand"]:
                    n += 1
        return n

    def _apply_filter_results(
        self,
        pickaxe: Pickaxe,
        compound_ids_to_check: List[str] = [],
        reaction_ids_to_delete: List[str] = [],
    ) -> None:
        """Apply filter results to Pickaxe object.

        Remove compounds and reactions that can be removed.
        For a compound to be removed it must:
            1. Not be flagged for expansion
            2. Not have a coproduct in a reaction marked for expansion
            3. Start with "C"

        Parameters
        ----------
        pickaxe : Pickaxe
            Instance of Pickaxe being used to expand and filter the network,
            this method modifies the Pickaxe object's compound documents.
        compound_ids_to_check : List[str]
            List of compound IDs to try to remove, if possible.
        """

        def should_delete_reaction(rxn_id: str) -> bool:
            """Returns whether or not a reaction can safely be deleted."""
            products = pickaxe.reactions[rxn_id]["Products"]
            for _, c_id in products:
                if c_id.startswith("C") and c_id not in cpds_to_remove:
                    return False
            # Every compound isn't in cpds_to_remove
            return True

        def remove_reaction(rxn_id):
            """Removes reaction and any resulting orphan compounds"""
            cpds_to_return = set()
            # Remove affiliations of reaction and check for orphans
            product_ids = [cpd[1] for cpd in pickaxe.reactions[rxn_id]["Products"]]
            for prod_id in product_ids:
                if prod_id.startswith("C"):
                    pickaxe.compounds[prod_id]["Product_of"].remove(rxn_id)
                    cpds_to_return.add(prod_id)
            compound_ids = [cpd[1] for cpd in pickaxe.reactions[rxn_id]["Reactants"]]
            for cpd_id in compound_ids:
                if cpd_id.startswith("C"):
                    pickaxe.compounds[cpd_id]["Reactant_in"].remove(rxn_id)
                    cpds_to_return.add(cpd_id)
            # Delete reaction itself
            del pickaxe.reactions[rxn_id]

            return cpds_to_return

        # Process reactions to delete
        # Loop through reactions to add compounds to check and to delete reactions
        if reaction_ids_to_delete:
            cpd_check_from_rxn = set()
            for rxn_id in reaction_ids_to_delete:
                cpd_check_from_rxn = cpd_check_from_rxn.union(remove_reaction(rxn_id))

            # Check for orphaned compounds due to reaction deletion
            while len(cpd_check_from_rxn) != 0:
                cpd_id = cpd_check_from_rxn.pop()
                # Orphan compound is one that has no reaction connecting it
                if cpd_id in pickaxe.compounds:
                    product_of = copy(pickaxe.compounds[cpd_id].get("Product_of", []))
                    # Delete if no reactions
                    if not product_of:
                        # Delete out reactions
                        reactant_in = copy(
                            pickaxe.compounds[cpd_id].get("Reactant_in", [])
                        )
                        for rxn_id in reactant_in:
                            cpd_check_from_rxn = cpd_check_from_rxn.union(
                                remove_reaction(rxn_id)
                            )
                        # Now delete compound
                        del pickaxe.compounds[cpd_id]

        # Go through compounds_ids_to_check and delete cpds/rxns as needed
        if compound_ids_to_check:
            cpds_to_remove = set()
            rxns_to_check = []

            compound_ids_to_check = set(compound_ids_to_check)
            for cpd_id in compound_ids_to_check:
                cpd_dict = pickaxe.compounds.get(cpd_id)
                if not cpd_dict:
                    continue

                if not cpd_dict["Expand"] and cpd_id.startswith("C"):
                    cpds_to_remove.add(cpd_id)

                    rxns_to_check.extend(pickaxe.compounds[cpd_id]["Product_of"])
                    rxns_to_check.extend(pickaxe.compounds[cpd_id]["Reactant_in"])

            rxns_to_check = set(rxns_to_check)
            # Function to check to see if should delete reaction
            # If reaction has compound that won't be deleted keep it
            # Check reactions for deletion
            for rxn_id in rxns_to_check:
                if should_delete_reaction(rxn_id):
                    for _, c_id in pickaxe.reactions[rxn_id]["Products"]:
                        if c_id.startswith("C"):
                            if rxn_id in pickaxe.compounds[c_id]["Product_of"]:
                                pickaxe.compounds[c_id]["Product_of"].remove(rxn_id)

                    for _, c_id in pickaxe.reactions[rxn_id]["Reactants"]:
                        if c_id.startswith("C"):
                            if rxn_id in pickaxe.compounds[c_id]["Reactant_in"]:
                                pickaxe.compounds[c_id]["Reactant_in"].remove(rxn_id)

                    del pickaxe.reactions[rxn_id]
                else:
                    # Reaction is dependent on compound that is flagged to be
                    # removed. Don't remove compound
                    products = pickaxe.reactions[rxn_id]["Products"]
                    cpds_to_remove -= set(i[1] for i in products)

                    # for _, c_id in products:
                    #     if c_id in cpds_to_remove:
                    #         cpds_to_remove -= {c_id}

            # Remove compounds and reactions if any found
            for cpd_id in cpds_to_remove:
                del pickaxe.compounds[cpd_id]


if __name__ == "__main__":
    pass
